#ifndef __KDKERNELDEFINES_H_INCLUDED__
#define __KDKERNELDEFINES_H_INCLUDED__

#include <stdio.h>

///////////////////////////////////////////////////////////////////////////////
//
// CUDA kernels defines
//

extern "C" void __mx__debug__line_set__(int line);
extern "C" void __mx__debug__file_set__(const char *file);
extern "C" void __fucked_error		   (const char *format, ...);
extern "C" void __mxLogger			   (const char *format, ...);
#define mxLogger					   __mx__debug__line_set__(__LINE__), __mx__debug__file_set__(__FILE__), __mxLogger
#define fucked_error			       __mx__debug__line_set__(__LINE__), __mx__debug__file_set__(__FILE__), __fucked_error

extern "C" int  __TMS_CU();

extern "C" void __TME_CU(const char *v);

#define MAX_RESPECT			 1000000.0f
#define MATH_EPSILON			0.0001f
#define REFL_EPSILON			  0.05f
#define M_PI    3.14159265358979323846f

#define TRIRAY_PRECALC_EDGES

// #define FAST_HALF_FLOATS

///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 _asflt3(const float4 &a)
{
 return make_float3(a.x, a.y, a.z);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator * (const float3 &a, const float3 &b)
{
 return make_float3(a.x*b.x, a.y*b.y, a.z*b.z);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator *= (float3 &b, float f)
{
 b.x *= f;
 b.y *= f;
 b.z *= f;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator * (float f, const float3 &v)
{
 return make_float3(v.x*f, v.y*f, v.z*f);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator * (const float3 &v, float f)
{
 return make_float3(v.x*f, v.y*f, v.z*f);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float4 operator * (const float4 &v, float f)
{
 return make_float4(v.x*f, v.y*f, v.z*f, v.w*f);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator *= (float3 &b, const float3 &a)
{
 b.x *= a.x;
 b.y *= a.y;
 b.z *= a.z;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator + (const float3 &a, const float3 &b)
{
 return make_float3(a.x+b.x, a.y+b.y, a.z+b.z);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator + (const float3 &a, const float b)
{
 return make_float3(a.x+b, a.y+b, a.z+b);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator += (float3 &b, const float3 &a)
{
 b.x += a.x;
 b.y += a.y;
 b.z += a.z;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator += (float4 &b, const float4 &a)
{
 b.x += a.x;
 b.y += a.y;
 b.z += a.z;
 b.w += a.w;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator += (float3 &b, float a)
{
 b.x += a;
 b.y += a;
 b.z += a;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator - (const float3 &a, const float3 &b)
{
 return make_float3(a.x-b.x, a.y-b.y, a.z-b.z);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator -= (float3 &b, const float3 &a)
{
 b.x -= a.x;
 b.y -= a.y;
 b.z -= a.z;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator -= (float3 &b, float a)
{
 b.x -= a;
 b.y -= a;
 b.z -= a;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator / (const float3 &v, const float3 &f)
{
 return make_float3(__fdividef(v.x, f.x), __fdividef(v.y, f.y), __fdividef(v.z, f.z));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator /= (float3 &b, const float3 &a)
{
 b.x = __fdividef(a.x, b.x);
 b.y = __fdividef(a.y, b.y);
 b.z = __fdividef(a.z, b.z);
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 operator / (const float3 &v, float f)
{
 float inv = __fdividef(1.0f, f);
 return v * inv;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void operator /= (float3 &b, float f)
{
 float inv = __fdividef(1.0f, f);
 b.x *= inv;
 b.y *= inv;
 b.z *= inv;
 return;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float dot(const float3 &a, const float3 &b)
{
 return a.x * b.x + a.y * b.y + a.z * b.z;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float dot(const float4 &a, const float4 &b)
{
 return a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float clamp(float f, float a, float b)
{
 return fmaxf(a, fminf(f, b));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 cross(const float3 &a, const float3 &b)
{
 return make_float3(a.y * b.z - a.z * b.y,
					a.z * b.x - a.x * b.z,
					a.x * b.y - a.y * b.x);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ void xch(float &a, float &b)
{
 float t = a;
 a = b;
 b = t;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 fminf(const float3 &a, const float3 &b)
{
 return make_float3(fminf(a.x, b.x),
					fminf(a.y, b.y),
					fminf(a.z, b.z));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 fmaxf(const float3 &a, const float3 &b)
{
 return make_float3(fmaxf(a.x, b.x),
					fmaxf(a.y, b.y),
					fmaxf(a.z, b.z));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float4 fminf(const float4 &a, const float4 &b)
{
 return make_float4(fminf(a.x, b.x),
					fminf(a.y, b.y),
					fminf(a.z, b.z),
					fminf(a.w, b.w));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float4 fmaxf(const float4 &a, const float4 &b)
{
 return make_float4(fmaxf(a.x, b.x),
					fmaxf(a.y, b.y),
					fmaxf(a.z, b.z),
					fmaxf(a.w, b.w));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 lerp(const float3 &x, const float3 &y, float s)
{
 return x + (s * (y - x));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 reflect(const float3 &Ray, const float3 &Normal)
{
 return Ray - (2.0f * Normal * dot(Normal, Ray));
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ unsigned char __float_as_uchar(float v)
{
 return __saturatef(v) * 255.0f;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float __uchar_as_float(unsigned char v)
{
 return (float)v * (1.0f / 255.0f);
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 rotateX(const float3 &v, float angle)
{
 float3 res;
 float  sina, cosa;
 __sincosf(angle, &sina, &cosa);

 res.x =  v.x;
 res.y = (v.y * cosa) - (v.z * sina);
 res.z = (v.y * sina) + (v.z * cosa);

 return res;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 rotateY(const float3 &v, float angle)
{
 float3 res;
 float  sina, cosa;
 __sincosf(angle, &sina, &cosa);

 res.x = (v.x * cosa) - (v.z * sina);
 res.y =  v.y;
 res.z = (v.x * sina) + (v.z * cosa);

 return res;
}
///////////////////////////////////////////////////////////////////////////////
inline __device__ float3 rotateZ(const float3 &v, float angle)
{
 float3 res;
 float  sina, cosa;
 __sincosf(angle, &sina, &cosa);

 res.x = (v.x * cosa) - (v.y * sina);
 res.y = (v.x * sina) + (v.y * cosa);
 res.z =  v.z;

 return res;
}
///////////////////////////////////////////////////////////////////////////////

#ifndef FAST_HALF_FLOATS
//
// Generic conversion, handles NANs, Infs and denormals
//
///////////////////////////////////////////////////////////////////////////////
__device__ unsigned short __float_as_half(float Value)
{
 //
 // Zero case
 //
 if (Value == 0.0f) 
	 return 0;

 //
 // Generic case
 //
 int i = __float_as_int(Value);
 int s =  (i >> 16) & 0x00008000;
 int e = ((i >> 23) & 0x000000ff) - (127 - 15);
 int m =   i        & 0x007fffff;

 if (e <= 0)
 {
	 //
	 // Below half_min
	 //
	 if (e < -10)
		 return 0;

     m = (m | 0x00800000) >> (1 - e);

     if (m &  0x00001000)
         m += 0x00002000;

     return s | (m >> 13);
 }
 else 
 if (e == 0xff - (127 - 15))
 {
	 if (m == 0)
	 {
         return s | 0x7c00;
	 }

     m >>= 13;
     
	 return s | 0x7c00 | m | (m == 0);
 }

 if (m & 0x00001000)
 {
     m += 0x00002000;

     if (m & 0x00800000)
     {
		 m =  0; 
         e += 1; 
     }
 }

 if (e > 30)
     return s | 0x7c00;

 return s | (e << 10) | (m >> 13);
}
///////////////////////////////////////////////////////////////////////////////
__device__ float __half_as_float(unsigned int Value)
{
 //
 // Zero case
 //
 if (Value == 0)
	 return 0.0f;

 //
 // Generic case
 //
 unsigned int s, e, m;

 m = (Value & 0x03FF);
 s = (Value & 0x8000);

 if ((Value & 0x7C00) != 0)
 {
	 e = ((Value >> 10) & 0x1F);
 }
 else 
 if (m != 0)
 {
     e = 1;

     do
     {
		 e--;
         m <<= 1;
     } while ((m & 0x0400) == 0);

     m &= 0x03FF;
 }
 else
 {
	 e = (unsigned int)-112;
 }

 return __int_as_float((s << 16) | ((e + 112) << 23) | (m << 13));
}
///////////////////////////////////////////////////////////////////////////////

#else
//
// Fast but not so generic conversion, ok to handle values <-1024.0, 1024.0>
// without denormals, NANs and Infs
//
///////////////////////////////////////////////////////////////////////////////
__device__ unsigned short __float_as_half(float Value)
{
 if (Value == 0.0f) 
	 return 0;

 int i = __float_as_int(Value);
 int s =  (i >> 16) & 0x00008000;
 int e = ((i >> 23) & 0x000000ff) - (127 - 15);
 int m =   i        & 0x007fffff;

 if (e < -10)
	 return 0;

 return s | (e << 10) | (m >> 13);
}
///////////////////////////////////////////////////////////////////////////////
__device__ float __half_as_float(unsigned int Value)
{
 if (Value == 0)
	 return 0.0f;

 unsigned int m = (Value & 0x03FF);
 unsigned int s = (Value & 0x8000);
 unsigned int e = ((Value >> 10) & 0x1F);

 return __int_as_float((s << 16) | ((e + 112) << 23) | (m << 13));
}
///////////////////////////////////////////////////////////////////////////////
#endif // FAST_HALF_FLOATS

#endif // __KDKERNELDEFINES_H_INCLUDED__
