#include "misc.h"
#include "vector.h"
#include "matrix.h"
#pragma  hdrstop

#pragma comment(lib, "d3d9.lib")
#pragma comment(lib, "d3dx9.lib")
#pragma comment(lib, "dxguid.lib")
#pragma comment(lib, "cuda.lib")
#pragma comment(lib, "cudart.lib")

extern "C" void CUDA_DFKernel(void        *Surface, 
							  unsigned int RTSizeX, 
							  unsigned int RTSizeY);

extern "C" void CUDA_BindConsts(unsigned int RTSizeX,
								unsigned int RTSizeY,
								unsigned int RTPitch,
								const float *RayOrigin,
								const float *ProjVector,
								const float *ModelviewInv);

#define PRINT_TIMINGS

extern "C" int  __TMS_CU();

extern "C" void __TME_CU(const char *v);

#ifdef PRINT_TIMINGS
static __int64 __Freq, __StartTimeCU, __EndTimeCU;
#endif

int __TMS_CU()
{
#ifdef PRINT_TIMINGS
 if (__Freq == 0)
	 QueryPerformanceFrequency((LARGE_INTEGER*)&__Freq);
 cudaThreadSynchronize(); // TMESync
 QueryPerformanceCounter((LARGE_INTEGER*)&__StartTimeCU);
#endif
 return 1;
}

void __TME_CU(const char *v)
{
 cudaThreadSynchronize(); // TMESync
#ifdef PRINT_TIMINGS
 QueryPerformanceCounter((LARGE_INTEGER*)&__EndTimeCU);
 // mxLogger("[%s] - [%f seconds]", v, (float)(__EndTimeCU - __StartTimeCU) / (float)__Freq);
#endif
 return;
}

///////////////////////////////////////////////////////////////////////////////

#define WINDOW_SIZE_X			640
#define WINDOW_SIZE_Y			480
#define CAMERA_FOV				60
#define WINDOW_CLASS_NAME "MX-DFRND"

static HINSTANCE		  ghInst;
static HWND				  ghWnd;
static DWORD			  gTotalNumFrames;
static DWORD			  gLastSecFrames;
static DWORD			  gLastSecTime;
static float			  gCurrFPS;
static c_VECTOR			  gCameraPosition = c_VECTOR(-142.967148f, -87.765739f, -11.932617f, 0.0f);
static c_VECTOR			  gCameraDirection;
static float			  gCameraX	  = 0.1f;
static float			  gCameraY	  = 0.7f;
static int				  gLastMouseX = 0xFFFF;
static int				  gLastMouseY = 0xFFFF;
static DWORD			  gDeltaTick;
static IDirect3D9		 *gD3D;
static IDirect3DDevice9	 *gD3DDevice;
static int				  gCudaDeviceID;
static cudaDeviceProp	  gCudaDeviceProps;
static IDirect3DTexture9 *gBackBuffer;

static void Initialize();
static void UpdateCamera();
static BOOL Update();
static void Deinitialize();

///////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
 switch (message) 
 {
	 case WM_DESTROY:
		 {
			 PostQuitMessage(0);
		 }
     return 0;
 }

 return DefWindowProc(hWnd, message, wParam, lParam);
}
///////////////////////////////////////////////////////////////////////////////
int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow)
{
#if _DEBUG
  _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
#endif

 __SetThreadName(-1, "MainThread");
 __SetThreadFPUCW();

 ghInst = hInstance;
 
 WNDCLASSEX wcex;

 wcex.cbSize		 = sizeof(WNDCLASSEX); 
 wcex.style          = CS_HREDRAW | CS_VREDRAW;
 wcex.lpfnWndProc    = WndProc;
 wcex.cbClsExtra     = 0;
 wcex.cbWndExtra     = 0;
 wcex.hInstance      = hInstance;
 wcex.hIcon          = NULL;
 wcex.hCursor        = LoadCursor(NULL, IDC_ARROW);
 wcex.hbrBackground  = (HBRUSH)(COLOR_WINDOW+1);
 wcex.lpszMenuName   = NULL;
 wcex.lpszClassName  = WINDOW_CLASS_NAME;
 wcex.hIconSm        = NULL;

 if (!RegisterClassEx(&wcex))
 {
	 fucked_error("Cannot register window class !");
	 return 0;
 }

 RECT rc = 
 { 
	 0, 
	 0, 
	 WINDOW_SIZE_X, 
	 WINDOW_SIZE_Y 
 };

 AdjustWindowRect(&rc, WS_OVERLAPPEDWINDOW, FALSE);

 ghWnd = CreateWindow(WINDOW_CLASS_NAME, 
					  WINDOW_CLASS_NAME, 
					  WS_OVERLAPPEDWINDOW,
                      0, 
					  0, 
					  rc.right - rc.left, 
					  rc.bottom - rc.top, 
					  NULL, 
					  NULL, 
					  hInstance, 
					  NULL);

 if (ghWnd == NULL)
 {
	 fucked_error("Cannot create window !");
	 return 0;
 }

 ShowWindow(ghWnd, nCmdShow);

 Initialize();

 MSG msg = { 0 };

#ifdef PRINT_TIMINGS
 __int64 QPF;
 QueryPerformanceFrequency((LARGE_INTEGER*)&QPF);
#endif

 while (WM_QUIT != msg.message)
 {
	 if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
     {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
     }
     else
     {
		 if ((::GetTickCount() - gLastSecTime) > 1000)
		 {
			 gCurrFPS       = (float)gLastSecFrames;
			 gLastSecTime   = ::GetTickCount();
			 gLastSecFrames = 0;
		 }

		 gTotalNumFrames++;
		 gLastSecFrames++;

#ifdef PRINT_TIMINGS
		 __int64 StartFrame, EndFrame;
		 QueryPerformanceCounter((LARGE_INTEGER*)&StartFrame);
#endif

		 if (Update() == FALSE)
			 break;

#ifdef PRINT_TIMINGS
		 QueryPerformanceCounter((LARGE_INTEGER*)&EndFrame);
		 mxLogger("Time of this frame: %f", (float)(EndFrame - StartFrame) / (float)QPF);
#endif

		 if ((gTotalNumFrames % 10) == 0)
		 {
			 char buff[MAX_FNAME];
			 sprintf_s(buff, sizeof(buff)/sizeof(buff[0]), "Frame: %i FPS: %f", gTotalNumFrames, gCurrFPS);
			 SetWindowText(ghWnd, buff);
		 }
     }
 }

 Deinitialize();

 UnregisterClass(WINDOW_CLASS_NAME, wcex.hInstance);
 mxLogger("All is done !");
 return 1;
}
///////////////////////////////////////////////////////////////////////////////
void Initialize()
{
 CoInitializeEx(NULL, COINIT_MULTITHREADED);

 gD3D = Direct3DCreate9(D3D_SDK_VERSION);
 if (gD3D == NULL)
	 fucked_error("Cannot create D3D9 interface !");

 DWORD AdapterID   = 0;
 BOOL  CudaEnabled = FALSE;

 for(; AdapterID < gD3D->GetAdapterCount(); AdapterID++)
 {
	 D3DADAPTER_IDENTIFIER9 ident;
     int device;

     gD3D->GetAdapterIdentifier(AdapterID, 0, &ident);
     cudaD3D9GetDevice(&device, ident.DeviceName);

     if (cudaSuccess == cudaGetLastError()) 
	 {
		 CudaEnabled   = TRUE;
		 gCudaDeviceID = device;
		 break;
	 }
 }

 if (CudaEnabled == FALSE)
	 fucked_error("Cuda enabled device not found !");

 D3DDISPLAYMODE d3ddm;
 if (FAILED(gD3D->GetAdapterDisplayMode(AdapterID, &d3ddm)))
	 fucked_error(NULL);

 D3DPRESENT_PARAMETERS d3dpp; 
 memset(&d3dpp, 0, sizeof(d3dpp));

 d3dpp.BackBufferWidth		  = WINDOW_SIZE_X;
 d3dpp.BackBufferHeight		  = WINDOW_SIZE_Y;
 d3dpp.Windowed				  = TRUE;
 d3dpp.MultiSampleType		  = D3DMULTISAMPLE_NONE;
 d3dpp.MultiSampleQuality	  = D3DMULTISAMPLE_NONE;
 d3dpp.Flags				  = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL | D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
 d3dpp.BackBufferFormat		  = D3DFMT_A8R8G8B8;
 d3dpp.SwapEffect			  = D3DSWAPEFFECT_DISCARD;
 d3dpp.EnableAutoDepthStencil = TRUE;
 d3dpp.AutoDepthStencilFormat = D3DFMT_D24X8;
 d3dpp.hDeviceWindow          = ghWnd;
 d3dpp.BackBufferCount		  = 2;
 d3dpp.PresentationInterval   = D3DPRESENT_INTERVAL_ONE;
 
 DWORD      bflags     = D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_MULTITHREADED | D3DCREATE_FPU_PRESERVE;
 D3DDEVTYPE DevType    = D3DDEVTYPE_HAL;
 DWORD      AdapterUID = AdapterID;

 HRESULT res = gD3D->CreateDevice(AdapterUID,
							      DevType, 
								  ghWnd,
								  bflags,
                                  &d3dpp, 
								  &gD3DDevice);

 if (FAILED(res))
	 fucked_error("Cannot create D3D device with specified flags !");

 D3DADAPTER_IDENTIFIER9 VendorID;
 gD3D->GetAdapterIdentifier(AdapterID, 0, &VendorID);

 mxLogger("Initializing D3D9 interface");
 mxLogger("Driver name : [%s]", VendorID.Driver);
 mxLogger("Driver desc : [%s]", VendorID.Description);
 mxLogger("Device name : [%s]", VendorID.DeviceName);
 mxLogger("Vendor ID   : [%i]", VendorID.VendorId);
 mxLogger("Device ID   : [%i]", VendorID.DeviceId);
 mxLogger("SubSys ID   : [%i]", VendorID.SubSysId);
 mxLogger("Revision    : [%i]", VendorID.Revision);
 mxLogger("WHQL        : [%X]", VendorID.WHQLLevel);

 if (cudaGetDeviceProperties(&gCudaDeviceProps, gCudaDeviceID) != cudaSuccess)
	 fucked_error("Error initializing Cuda !");

 if (gCudaDeviceID == 0) 
 {
	 if (gCudaDeviceProps.major == 9999 && gCudaDeviceProps.minor == 9999)
		 fucked_error("Error initializing Cuda !");
 }

 mxLogger("Initializing Cuda D3D9 interface");
 mxLogger("Device       %i: [%s] ",		   gCudaDeviceID, gCudaDeviceProps.name);
 mxLogger("Revision number: %i.%i",		   gCudaDeviceProps.major, gCudaDeviceProps.minor);
 mxLogger("Global memory  : %i   ",		   (int)gCudaDeviceProps.totalGlobalMem);
 mxLogger("Multiprocessors: %i   ",		   gCudaDeviceProps.multiProcessorCount);
 mxLogger("Number of cores: %i   ",		   8 * gCudaDeviceProps.multiProcessorCount);
 mxLogger("Constant memory: %i   ",		   gCudaDeviceProps.totalConstMem); 
 mxLogger("Shared memory  : %i   ",		   gCudaDeviceProps.sharedMemPerBlock);
 mxLogger("Registers avail: %i   ",		   gCudaDeviceProps.regsPerBlock);
 mxLogger("Max block size : %i x %i x %i", (int)gCudaDeviceProps.maxThreadsDim[0], (int)gCudaDeviceProps.maxThreadsDim[1], (int)gCudaDeviceProps.maxThreadsDim[2]);
 mxLogger("Max grid size  : %i x %i x %i", (int)gCudaDeviceProps.maxGridSize[0], (int)gCudaDeviceProps.maxGridSize[1], (int)gCudaDeviceProps.maxGridSize[2]);
 mxLogger("Clock rate     : %.2f GHz",     gCudaDeviceProps.clockRate * 1e-6f);
 mxLogger("Concurrent copy: %s",           gCudaDeviceProps.deviceOverlap ? "Yes" : "No");

 if (cudaD3D9SetDirect3DDevice(gD3DDevice) != cudaSuccess)
	 fucked_error("Error initializing Cuda !");

 mxLogger("D3D9 & Cuda initialized");

 if (FAILED(gD3DDevice->CreateTexture(WINDOW_SIZE_X, WINDOW_SIZE_Y, 1, D3DUSAGE_DYNAMIC, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &gBackBuffer, NULL)))
	 fucked_error("Cannot create Cuda interop texture !");

#ifndef _DEBUG
 if (cudaD3D9RegisterResource(gBackBuffer, cudaD3D9RegisterFlagsNone) != cudaSuccess)
	 fucked_error("Cannot register Cuda resource !");

 if (cudaD3D9ResourceSetMapFlags(gBackBuffer, cudaD3D9MapFlagsWriteDiscard) != cudaSuccess)
	 fucked_error("Cannot set Cuda flags !");
#endif

 return;
}
///////////////////////////////////////////////////////////////////////////////
void UpdateCamera()
{
 POINT CursorPos;
 GetCursorPos(&CursorPos);

 if (gLastMouseX == 0xFFFF)
 {
	 gLastMouseX = CursorPos.x;
	 gLastMouseY = CursorPos.y;
 }

 int CursorDeltaX = CursorPos.x - gLastMouseX;
 int CursorDeltaY = CursorPos.y - gLastMouseY;

 BOOL MoveForward  = FALSE;
 BOOL MoveBackward = FALSE;
 BOOL MoveLeft     = FALSE;
 BOOL MoveRight	   = FALSE;

 if (GetForegroundWindow() == ghWnd)
 {
	 SetCursorPos(WINDOW_SIZE_X/2, WINDOW_SIZE_Y/2);

	 if ((GetKeyState(0x57) & 0x8000) == 0x8000) MoveForward  = TRUE;
	 if ((GetKeyState(0x53) & 0x8000) == 0x8000) MoveBackward = TRUE;
	 if ((GetKeyState(0x41) & 0x8000) == 0x8000) MoveLeft     = TRUE;
	 if ((GetKeyState(0x44) & 0x8000) == 0x8000) MoveRight    = TRUE;
 }
 else
 {
	 gLastMouseX = 0xFFFF;
	 gLastMouseY = 0xFFFF;
	 gDeltaTick  = 0;
 }

 gLastMouseX = WINDOW_SIZE_X/2;
 gLastMouseY = WINDOW_SIZE_Y/2;

 gCameraX -= float(CursorDeltaX) * 0.006f;
 gCameraY -= float(CursorDeltaY) * 0.006f;

 c_MATRIX RotY, RotZ, Final;
 
 RotY.MakeRotation(gCameraX, 0.0, 1.0, 0.0);
 RotZ.MakeRotation(gCameraY, 0.0, 0.0, 1.0);
 Final = RotY * RotZ;

 gCameraDirection = c_VECTOR(1.0f, 0.0f, 0.0f, 0.0f);
 Final.MullNorm3(&gCameraDirection, &gCameraDirection);

 DWORD CurrTick = GetTickCount();

 if (gDeltaTick == 0)
	 gDeltaTick = CurrTick;

 float Delta = (float(CurrTick - gDeltaTick) / 1000.0f) * 60.0f;

 gDeltaTick = CurrTick;

 if (MoveForward)
	 gCameraPosition += gCameraDirection * Delta;

 if (MoveBackward)
	 gCameraPosition -= gCameraDirection * Delta;

 if (MoveLeft)
 {
	 c_VECTOR Str = c_VECTOR(0.0f, 0.0f, 1.0f, 0.0f);
	 Final.MullNorm3(&Str, &Str);
	 gCameraPosition -= Str * Delta;
 }

 if (MoveRight)
 {
	 c_VECTOR Str = c_VECTOR(0.0f, 0.0f, 1.0f, 0.0f);
	 Final.MullNorm3(&Str, &Str);
	 gCameraPosition += Str * Delta;
 }
 
 return;
}
///////////////////////////////////////////////////////////////////////////////
BOOL Update()
{
 //
 // Update camera
 //
 UpdateCamera();

 //
 // Prepare camera matrix
 //
 c_MATRIX CameraMatrix, ProjectionMatrix;
 
 ProjectionMatrix.MakeProjectionMatrix(CAMERA_FOV, (float)WINDOW_SIZE_X / (float)WINDOW_SIZE_Y, 1.0, 1024.0);

 CameraMatrix.MakeLookAtMatrix(gCameraPosition.v.x, 
							   gCameraPosition.v.y, 
							   gCameraPosition.v.z,
							   gCameraPosition.v.x + gCameraDirection.v.x, 
							   gCameraPosition.v.y + gCameraDirection.v.y, 
							   gCameraPosition.v.z + gCameraDirection.v.z, 
							   0.0f,
							   1.0f,
							   0.0f);

 CameraMatrix.Inverse();

 //
 // Map resource
 //
#ifndef _DEBUG
 if (cudaD3D9MapResources(1, (IDirect3DResource9 **)&gBackBuffer) != cudaSuccess)
	 fucked_error("Cannot map Cuda resource !");
#endif

 void *TexData;
 unsigned int TexPitch;

#ifndef _DEBUG
 cudaD3D9ResourceGetMappedPointer(&TexData, gBackBuffer, 0, 0);
 cudaD3D9ResourceGetMappedPitch  (&TexPitch, NULL, gBackBuffer, 0, 0);
#else
 D3DLOCKED_RECT lr;
 if (FAILED(gBackBuffer->LockRect(0, &lr, NULL, D3DLOCK_DISCARD)))
	 fucked_error(NULL);
 TexData = lr.pBits;
 TexPitch= lr.Pitch;
#endif

 const float InvXHalf = 2.0f / (float)WINDOW_SIZE_X;
 const float InvYHalf = 2.0f / (float)WINDOW_SIZE_Y;
 const float InvM_00  = 1.0f / ProjectionMatrix.m._00;
 const float InvM_11  =-1.0f / ProjectionMatrix.m._11;

 c_VECTOR ProjVector = c_VECTOR(InvXHalf, InvYHalf, InvM_00, InvM_11);

 //
 // Invoke kernel
 //
 CUDA_BindConsts(WINDOW_SIZE_X,
				 WINDOW_SIZE_Y,
				 TexPitch,
				 gCameraPosition.v.v,
				 ProjVector.v.v,
				 CameraMatrix.m.v);

 CUDA_DFKernel(TexData, WINDOW_SIZE_X, WINDOW_SIZE_Y);

 //
 // Unmap resource
 //
#ifndef _DEBUG
 if (cudaD3D9UnmapResources(1, (IDirect3DResource9 **)&gBackBuffer) != cudaSuccess)
	 fucked_error("Cannot unmap Cuda resource !");
#else
 gBackBuffer->UnlockRect(0);
#endif

 //
 // Show the result
 //
 const unsigned int IB[6] = { 0, 1, 2, 0, 2, 3 };

 #pragma pack(1)
 struct FVFVertex
 {
	 float p[3];
	 float t[2];
 };
 #pragma pack()

 gD3DDevice->BeginScene();
 gD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET, 0, 1.0f, 0);	
 gD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
 gD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
 gD3DDevice->SetFVF(D3DFVF_XYZ|D3DFVF_TEX1|D3DFVF_TEXCOORDSIZE2(0));

 const float tsx = (1.0f / (float)WINDOW_SIZE_X) * 0.5f;
 const float tsy = (1.0f / (float)WINDOW_SIZE_Y) * 0.5f;

 const FVFVertex VB[4] = 
 {
	 { {-1.0f, -1.0f, 0.0f }, { 0.0f + tsx, 0.0f + tsy, } },
 	 { { 1.0f, -1.0f, 0.0f }, { 1.0f + tsx, 0.0f + tsy, } },
	 { { 1.0f,  1.0f, 0.0f }, { 1.0f + tsx, 1.0f + tsy, } },
	 { {-1.0f,  1.0f, 0.0f }, { 0.0f + tsx, 1.0f + tsy, } }
 };
 
 gD3DDevice->SetTexture(0, gBackBuffer);
 gD3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
 gD3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_POINT);
 gD3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU,  D3DTADDRESS_CLAMP);
 gD3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV,  D3DTADDRESS_CLAMP);

 gD3DDevice->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, 4, 2, IB, D3DFMT_INDEX32, VB, sizeof(FVFVertex));
 gD3DDevice->SetTexture(0, NULL);

 gD3DDevice->EndScene();

 gD3DDevice->Present(NULL, NULL, NULL, NULL);

 //
 // Exit on device errors
 //
 if (gD3DDevice->TestCooperativeLevel() == D3DERR_DEVICELOST || gD3DDevice->TestCooperativeLevel() == D3DERR_DEVICEHUNG)
	 return FALSE;

 return TRUE;
}
///////////////////////////////////////////////////////////////////////////////
void Deinitialize()
{
 if (gBackBuffer)
 {
#ifndef _DEBUG
	 cudaD3D9UnregisterResource(gBackBuffer);
#endif
	 gBackBuffer->Release();
 }

 if (gD3DDevice)
	 gD3DDevice->Release();

 if (gD3D)
	 gD3D->Release();

 CoUninitialize();

 return;
}
///////////////////////////////////////////////////////////////////////////////
