#include "misc.h"
#include "vector.h"
#include "matrix.h"
#pragma  hdrstop

const c_MATRIX gc_MATRIX_Identity;

static __m128  __xmmONE  = { 1.000f, 1.000f, 1.000f, 1.000f };
static __m128  __xmmMASK = { 1.000f, 1.000f, 1.000f, 0.000f };
static __m128  __xmmTSSK = { 0.000f, 0.000f, 0.000f, 1.000f };
static __m128  __xmmEPS  = { 0.001f, 0.001f, 0.001f, 0.001f };
static __m128  __xmmQUAD = { 0.250f, 0.250f, 0.250f, 0.250f };
static __m128  __xmmSGNM = { 1.000f,-1.000f,-1.000f, 1.000f };

///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::MakeProjectionMatrix(float FOV, float Aspect, float NearPlane, float FarPlane)
{
 float cot_fov = 1.0f / tanf(FOV*(M_PI/360.0f));
 
 float fcx = cot_fov / Aspect;
 float fcy = cot_fov;
 float a   = FarPlane / (NearPlane-FarPlane);
 float b   = NearPlane*FarPlane/(NearPlane-FarPlane);

 #define M(col, row) m.v[(col<<2)+row]

 M(0,0) = fcx;  M(0,1) = 0.0f; M(0,2) = 0; M(0,3) = 0.0f;
 M(1,0) = 0.0f; M(1,1) = fcy;  M(1,2) = 0; M(1,3) = 0.0f;
 M(2,0) = 0.0f; M(2,1) = 0.0f; M(2,2) = a; M(2,3) =-1.0f;
 M(3,0) = 0.0f; M(3,1) = 0.0f; M(3,2) = b; M(3,3) = 0.0f;

 #undef M

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::MakeLookAtMatrix(float eyex,    float eyey,    float eyez,
		                        float centerx, float centery, float centerz,
						        float upx,     float upy,     float upz)
{
 D3DXVECTOR3 eye, at, up;

 eye.x = eyex;    eye.y = eyey;    eye.z = eyez;
 at.x  = centerx; at.y  = centery; at.z  = centerz;
 up.x  = upx;     up.y  = upy;     up.z  = upz;

 D3DXMatrixLookAtRH((D3DXMATRIX *)this, &eye, &at, &up);

 /*
 float x[3], y[3], z[3];
 float mag;

 z[0] = eyex - centerx;
 z[1] = eyey - centery;
 z[2] = eyez - centerz;

 mag = 1.0 / sqrtf(z[0]*z[0] + z[1]*z[1] + z[2]*z[2]);
 z[0] *= mag;
 z[1] *= mag;
 z[2] *= mag;

 y[0] = upx;
 y[1] = upy;
 y[2] = upz;

 x[0] =  y[1]*z[2] - y[2]*z[1];
 x[1] = -y[0]*z[2] + y[2]*z[0];
 x[2] =  y[0]*z[1] - y[1]*z[0];

 y[0] =  z[1]*x[2] - z[2]*x[1];
 y[1] = -z[0]*x[2] + z[2]*x[0];
 y[2] =  z[0]*x[1] - z[1]*x[0];

 mag = 1.0 / sqrtf(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
 x[0] *= mag;
 x[1] *= mag;
 x[2] *= mag;

 mag = 1.0 / sqrtf(y[0]*y[0] + y[1]*y[1] + y[2]*y[2]);
 y[0] *= mag;
 y[1] *= mag;
 y[2] *= mag;

 #define M(row,col)  m.v[(col<<2)+row]

 M(0,0) = x[0];		M(0,1) = x[1];	  M(0,2) = x[2];	M(0,3) = 0.0f;
 M(1,0) = y[0];		M(1,1) = y[1];	  M(1,2) = y[2];	M(1,3) = 0.0f;
 M(2,0) = z[0];		M(2,1) = z[1];	  M(2,2) = z[2];	M(2,3) = 0.0f;
 M(3,0) = 0.0f;		M(3,1) = 0.0f;    M(3,2) = 0.0f;    M(3,3) = 1.0f;

 #undef M

 //
 // Translate eye to origin
 //
 c_MATRIX temp;
 temp.m.v[12] = -eyex;
 temp.m.v[13] = -eyex;
 temp.m.v[14] = -eyex;

 Mull(this, &temp);
 */

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::MakeOrthoMatrix(float l, float r, float b, float t,
							   float n, float f)
{
 Identity();

 m.v[0]  = 2.0f / (r - l);
 m.v[5]  = 2.0f / (t - b);
 m.v[10] = - (2.0f / (f - n));

 m.v[12] = - ((r + l) / (r - l));
 m.v[13] = - ((t + b) / (t - b));
 m.v[14] = - ((f + n) / (f - n));

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::ComposeFromMatrix(const c_MATRIX &mat)
{
 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6;

 _xmm5 = _mm_loadu_ps  (&mat.m.v[0]); // Load scale
 _xmm0 = _mm_loadu_ps  (&mat.m.v[4]); // Load quaternion
 _xmm6 = _mm_loadu_ps  (&mat.m.v[12]);// Load translation

 _xmm5 = _mm_mul_ps    (_xmm5, __xmmMASK); // Mask scale
 _xmm6 = _mm_mul_ps    (_xmm6, __xmmMASK); // Mask translate
 _xmm6 = _mm_add_ps    (_xmm6, __xmmTSSK); // Mask translate

 _xmm1 = _mm_mul_ps    (_xmm0, _xmm0); 

 _xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm2 = _mm_mul_ps    (_xmm2, _xmm0);

 _xmm3 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 1, 1));
 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 2));
 _xmm3 = _mm_mul_ps    (_xmm3, _xmm4);

 _xmm0 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(1, 2, 1, 2));
 _xmm0 = _mm_mul_ps    (_xmm0, __xmmSGNM);
 _xmm4 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 1, 2, 1));
 _xmm4 = _mm_add_ps    (_xmm4, _xmm0);      // y0, z0, x1, x2

 _xmm0 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 1, 0));
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 2, 2));
 _xmm0 = _mm_add_ps    (_xmm0, _xmm1);      // y1, x0, z2

 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(3, 3, 3, 3));
 _xmm2 = _mm_mul_ps    (_xmm2, __xmmSGNM);
 _xmm3 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm3 = _mm_add_ps    (_xmm3, _xmm2);      // z1, y2

 _xmm0 = _mm_add_ps    (_xmm0, _xmm0);      // y1, x0, z2     | * 2
 _xmm4 = _mm_add_ps    (_xmm4, _xmm4);      // y0, z0, x1, x2 | * 2
 _xmm3 = _mm_add_ps    (_xmm3, _xmm3);      // z1, y2		  | * 2
 _xmm0 = _mm_sub_ps    (__xmmONE, _xmm0);   // y1, x0, z2     | 1 - 

 //
 // Swizzle output
 //
 _xmm1 = _mm_movelh_ps (_xmm0, _xmm4);
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 3, 2, 1));

 _xmm3 = _mm_movelh_ps (_xmm3, _xmm0);
 _xmm2 = _mm_shuffle_ps(_xmm3, _xmm4, _MM_SHUFFLE(2, 2, 2, 0));
 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(0, 0, 1, 2));

 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm3, _MM_SHUFFLE(1, 1, 3, 3));
 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm4 = _mm_move_ss   (_xmm4, _xmm0);
 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(0, 0, 2, 1));

 _xmm5 = _mm_mul_ps    (_xmm5, __xmmMASK); // Mask scale
 _xmm6 = _mm_mul_ps    (_xmm6, __xmmMASK); // Mask translate
 _xmm6 = _mm_add_ps    (_xmm6, __xmmTSSK); // Mask translate

 _xmm1 = _mm_mul_ps(_xmm1, _xmm5);
 _xmm2 = _mm_mul_ps(_xmm2, _xmm5);
 _xmm4 = _mm_mul_ps(_xmm4, _xmm5);

 _mm_storeu_ps		   (&m.v[0],  _xmm1);
 _mm_storeu_ps		   (&m.v[4],  _xmm2);
 _mm_storeu_ps		   (&m.v[8],  _xmm4);
 _mm_storeu_ps		   (&m.v[12], _xmm6);	

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::LerpFromDecomposed(const c_MATRIX &a, const c_MATRIX &b, float alpha)
{
 //
 // Both matrices are already decomposed, just lerp'em
 // and compose result into standard matrix
 //
 // v[0-3]   -> Scale
 // v[4-7]   -> Quat rotation
 // v[8-11]  -> Unused
 // v[12-15] -> Translation
 //
 if (alpha < MATH_EPSILON)
 {
	 this->ComposeFromMatrix(a);
	 return;
 }
 else
 if (alpha > (1.0 - MATH_EPSILON))
 {
	 this->ComposeFromMatrix(b);
	 return;
 }

 __m128 Rot0, Rot1, Scl0, Scl1, Trns0, Trns1;

 Scl0  = _mm_loadu_ps(&a.m.v[0]);
 Rot0  = _mm_loadu_ps(&a.m.v[4]);
 Trns0 = _mm_loadu_ps(&a.m.v[12]);

 Scl1  = _mm_loadu_ps(&b.m.v[0]);
 Rot1  = _mm_loadu_ps(&b.m.v[4]);
 Trns1 = _mm_loadu_ps(&b.m.v[12]);

 __m128 alp = _mm_load_ss(&alpha);
 alp = _mm_shuffle_ps(alp, alp, _MM_SHUFFLE(0, 0, 0, 0));

 __m128 tmp = _mm_sub_ps(Scl1, Scl0);
		tmp = _mm_mul_ps(tmp, alp);
	   Scl0 = _mm_add_ps(Scl0, tmp);
	   Scl0 = _mm_mul_ps(Scl0, __xmmMASK);

	    tmp = _mm_sub_ps(Trns1, Trns0);
	    tmp = _mm_mul_ps(tmp, alp);
	  Trns0 = _mm_add_ps(Trns0, tmp);
	  Trns0 = _mm_mul_ps(Trns0, __xmmMASK);
	  Trns0 = _mm_add_ps(Trns0, __xmmTSSK);

 ((c_VECTOR*)&Rot0)->Slerp(*((c_VECTOR*)&Rot0), *((c_VECTOR*)&Rot1), alpha);

 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4;

 _xmm0 = Rot0;

 _xmm1 = _mm_mul_ps    (_xmm0, _xmm0); 

 _xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm2 = _mm_mul_ps    (_xmm2, _xmm0);

 _xmm3 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 1, 1));
 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 2));
 _xmm3 = _mm_mul_ps    (_xmm3, _xmm4);

 _xmm0 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(1, 2, 1, 2));
 _xmm0 = _mm_mul_ps    (_xmm0, __xmmSGNM);
 _xmm4 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 1, 2, 1));
 _xmm4 = _mm_add_ps    (_xmm4, _xmm0);      // y0, z0, x1, x2

 _xmm0 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 1, 0));
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 2, 2));
 _xmm0 = _mm_add_ps    (_xmm0, _xmm1);      // y1, x0, z2

 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(3, 3, 3, 3));
 _xmm2 = _mm_mul_ps    (_xmm2, __xmmSGNM);
 _xmm3 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm3 = _mm_add_ps    (_xmm3, _xmm2);      // z1, y2

 _xmm0 = _mm_add_ps    (_xmm0, _xmm0);      // y1, x0, z2     | * 2
 _xmm4 = _mm_add_ps    (_xmm4, _xmm4);      // y0, z0, x1, x2 | * 2
 _xmm3 = _mm_add_ps    (_xmm3, _xmm3);      // z1, y2		  | * 2
 _xmm0 = _mm_sub_ps    (__xmmONE, _xmm0);   // y1, x0, z2     | 1 - 

 //
 // Swizzle output
 //
 _xmm1 = _mm_movelh_ps (_xmm0, _xmm4);
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 3, 2, 1));

 _xmm3 = _mm_movelh_ps (_xmm3, _xmm0);
 _xmm2 = _mm_shuffle_ps(_xmm3, _xmm4, _MM_SHUFFLE(2, 2, 2, 0));
 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(0, 0, 1, 2));

 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm3, _MM_SHUFFLE(1, 1, 3, 3));
 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm4 = _mm_move_ss   (_xmm4, _xmm0);
 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(0, 0, 2, 1));

 _xmm1 = _mm_mul_ps(_xmm1, Scl0);
 _xmm2 = _mm_mul_ps(_xmm2, Scl0);
 _xmm4 = _mm_mul_ps(_xmm4, Scl0);

 _mm_storeu_ps		   (&m.v[0],  _xmm1);
 _mm_storeu_ps		   (&m.v[4],  _xmm2);
 _mm_storeu_ps		   (&m.v[8],  _xmm4);
 _mm_storeu_ps		   (&m.v[12], Trns0);

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::Lerp(const c_MATRIX &a, const c_MATRIX &b, float alpha)
{
 if (alpha < MATH_EPSILON)
 {
	 *this = a;
	 return;
 }
 else
 if (alpha > (1.0 - MATH_EPSILON))
 {
	 *this = b;
	 return;
 }

 __m128 Rot0, Rot1, Scl0, Scl1, Trns0, Trns1;

 a.Decompose((c_VECTOR*)&Scl0, (c_VECTOR*)&Rot0, (c_VECTOR*)&Trns0);
 b.Decompose((c_VECTOR*)&Scl1, (c_VECTOR*)&Rot1, (c_VECTOR*)&Trns1);

 __m128 alp = _mm_load_ss(&alpha);
 alp = _mm_shuffle_ps(alp, alp, _MM_SHUFFLE(0, 0, 0, 0));

 __m128 tmp = _mm_sub_ps(Scl1, Scl0);
		tmp = _mm_mul_ps(tmp, alp);
	   Scl0 = _mm_add_ps(Scl0, tmp);
	   Scl0 = _mm_mul_ps(Scl0, __xmmMASK);

	    tmp = _mm_sub_ps(Trns1, Trns0);
	    tmp = _mm_mul_ps(tmp, alp);
	  Trns0 = _mm_add_ps(Trns0, tmp);
	  Trns0 = _mm_mul_ps(Trns0, __xmmMASK);
	  Trns0 = _mm_add_ps(Trns0, __xmmTSSK);

 ((c_VECTOR*)&Rot0)->Slerp(*((c_VECTOR*)&Rot0), *((c_VECTOR*)&Rot1), alpha);

 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4;

 _xmm0 = Rot0;

 _xmm1 = _mm_mul_ps    (_xmm0, _xmm0); 

 _xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm2 = _mm_mul_ps    (_xmm2, _xmm0);

 _xmm3 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 1, 1));
 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 2));
 _xmm3 = _mm_mul_ps    (_xmm3, _xmm4);

 _xmm0 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(1, 2, 1, 2));
 _xmm0 = _mm_mul_ps    (_xmm0, __xmmSGNM);
 _xmm4 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 1, 2, 1));
 _xmm4 = _mm_add_ps    (_xmm4, _xmm0);      // y0, z0, x1, x2

 _xmm0 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 1, 0));
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 2, 2));
 _xmm0 = _mm_add_ps    (_xmm0, _xmm1);      // y1, x0, z2

 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(3, 3, 3, 3));
 _xmm2 = _mm_mul_ps    (_xmm2, __xmmSGNM);
 _xmm3 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm3 = _mm_add_ps    (_xmm3, _xmm2);      // z1, y2

 _xmm0 = _mm_add_ps    (_xmm0, _xmm0);      // y1, x0, z2     | * 2
 _xmm4 = _mm_add_ps    (_xmm4, _xmm4);      // y0, z0, x1, x2 | * 2
 _xmm3 = _mm_add_ps    (_xmm3, _xmm3);      // z1, y2		  | * 2
 _xmm0 = _mm_sub_ps    (__xmmONE, _xmm0);   // y1, x0, z2     | 1 - 

 //
 // Swizzle output
 //
 _xmm1 = _mm_movelh_ps (_xmm0, _xmm4);
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 3, 2, 1));

 _xmm3 = _mm_movelh_ps (_xmm3, _xmm0);
 _xmm2 = _mm_shuffle_ps(_xmm3, _xmm4, _MM_SHUFFLE(2, 2, 2, 0));
 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(0, 0, 1, 2));

 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm3, _MM_SHUFFLE(1, 1, 3, 3));
 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm4 = _mm_move_ss   (_xmm4, _xmm0);
 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(0, 0, 2, 1));

 _xmm1 = _mm_mul_ps(_xmm1, Scl0);
 _xmm2 = _mm_mul_ps(_xmm2, Scl0);
 _xmm4 = _mm_mul_ps(_xmm4, Scl0);

 _mm_storeu_ps		   (&m.v[0],  _xmm1);
 _mm_storeu_ps		   (&m.v[4],  _xmm2);
 _mm_storeu_ps		   (&m.v[8],  _xmm4);
 _mm_storeu_ps		   (&m.v[12], Trns0);

 return;
}
///////////////////////////////////////////////////////////////////////////////	
BOOL c_MATRIX::DecomposeInplace()
{
 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6, _xmm7;
 float AlReg[4] = { 0 };

 c_VECTOR *outScale  = (c_VECTOR*)(&m.v[0]);
 c_VECTOR *outRotate = (c_VECTOR*)(&m.v[4]);

 _xmm0 = _mm_loadu_ps(&m.v[ 0]);
 _xmm1 = _mm_loadu_ps(&m.v[ 4]);
 _xmm2 = _mm_loadu_ps(&m.v[ 8]);

 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm1, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm5 = _mm_shuffle_ps(_xmm5, _xmm2, _MM_SHUFFLE(0, 0, 2, 0));

 _xmm4 = _mm_mul_ps(_xmm5, _xmm5);
 _xmm5 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm6 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(2, 2, 2, 2));

 _xmm4 = _mm_add_ss(_xmm4, _xmm5);
 _xmm4 = _mm_add_ss(_xmm4, _xmm6);

 _xmm4 = _mm_rsqrt_ss(_xmm4);

 _mm_store_ss(&AlReg[0], _xmm4);

 _xmm4 = _mm_rcp_ss(_xmm4);

 // _xmm4.x = scaleX

 _mm_store_ss(&outScale->v.x, _xmm4);

 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm1, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm5 = _mm_shuffle_ps(_xmm5, _xmm2, _MM_SHUFFLE(1, 1, 2, 0));

 _xmm4 = _mm_mul_ps(_xmm5, _xmm5);
 _xmm5 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm6 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(2, 2, 2, 2));

 _xmm4 = _mm_add_ss(_xmm4, _xmm5);
 _xmm4 = _mm_add_ss(_xmm4, _xmm6);

 _xmm4 = _mm_rsqrt_ss(_xmm4);

 _mm_store_ss(&AlReg[1], _xmm4);

 _xmm4 = _mm_rcp_ss(_xmm4);

 // _xmm4.x = scaleY

 _mm_store_ss(&outScale->v.y, _xmm4);

 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm5 = _mm_shuffle_ps(_xmm5, _xmm2, _MM_SHUFFLE(2, 2, 2, 0));

 _xmm4 = _mm_mul_ps(_xmm5, _xmm5);
 _xmm5 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm6 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(2, 2, 2, 2));

 _xmm4 = _mm_add_ss(_xmm4, _xmm5);
 _xmm4 = _mm_add_ss(_xmm4, _xmm6);

 _xmm4 = _mm_rsqrt_ss(_xmm4);

 _mm_store_ss(&AlReg[2], _xmm4);

 _xmm4 = _mm_rcp_ss(_xmm4);
 _xmm7 = _mm_loadu_ps(AlReg);

 // _xmm4.x = scaleZ

 _mm_store_ss(&outScale->v.z, _xmm4);

 _xmm0 = _mm_mul_ps(_xmm0, _xmm7);
 _xmm1 = _mm_mul_ps(_xmm1, _xmm7);
 _xmm2 = _mm_mul_ps(_xmm2, _xmm7);

 //
 // _xmm[0-2] Normalized rotation matrix that need to be converted to QUAT
 //
 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0)); // xmm0.xxxx -> xmm4
 _xmm5 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1)); // xmm1.yyyy -> xmm5
 _xmm6 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 2, 2, 2)); // xmm2.zzzz -> xmm6

 if (_mm_comigt_ss(_xmm4, _xmm5) && _mm_comigt_ss(_xmm4, _xmm6))
 {
	 _xmm7 = _mm_add_ss(_xmm5,    _xmm6);
	 _xmm7 = _mm_sub_ss(_xmm4,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 *outRotate = c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
		 return TRUE;
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7); 
	 _xmm7 = _mm_rcp_ss(_xmm7); 
	 _mm_store_ss(&outRotate->v.x, _xmm3);

	 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm4 = _mm_add_ss(_xmm4, _xmm1);
	 _xmm4 = _mm_mul_ss(_xmm4, _xmm7);
	 _mm_store_ss(&outRotate->v.y, _xmm4);

	 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm2);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);	
	 _mm_store_ss(&outRotate->v.z, _xmm5);

	 _xmm6 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm6 = _mm_sub_ss(_xmm6, _xmm2);
	 _xmm6 = _mm_mul_ss(_xmm6, _xmm7);
	 _mm_store_ss(&outRotate->v.w, _xmm6);
 } 
 else 
 if (_mm_comigt_ss(_xmm5, _xmm6))
 {
	 _xmm7 = _mm_add_ss(_xmm4,    _xmm6);
	 _xmm7 = _mm_sub_ss(_xmm5,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 *outRotate = c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
		 return TRUE;
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7);
	 _xmm7 = _mm_rcp_ss(_xmm7);             
	 _mm_store_ss(&outRotate->v.y, _xmm3);

	 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm1);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);
	 _mm_store_ss(&outRotate->v.x, _xmm5);

	 _xmm6 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm3 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm6 = _mm_add_ss(_xmm6, _xmm3);
	 _xmm6 = _mm_mul_ss(_xmm6, _xmm7);
	 _mm_store_ss(&outRotate->v.z, _xmm6);

	 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm2 = _mm_sub_ss(_xmm2, _xmm0);
	 _xmm2 = _mm_mul_ss(_xmm2, _xmm7);
	 _mm_store_ss(&outRotate->v.w, _xmm2);
 } 
 else
 {
	 _xmm7 = _mm_add_ss(_xmm4,    _xmm5);
	 _xmm7 = _mm_sub_ss(_xmm6,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 *outRotate = c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
		 return TRUE;
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7);
	 _xmm7 = _mm_rcp_ss(_xmm7); 
	 _mm_store_ss(&outRotate->v.z, _xmm3);

	 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm4 = _mm_add_ss(_xmm4, _xmm2);
	 _xmm4 = _mm_mul_ss(_xmm4, _xmm7);
	 _mm_store_ss(&outRotate->v.x, _xmm4);

	 _xmm5 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm6 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm6);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);
	 _mm_store_ss(&outRotate->v.y, _xmm5);

	 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm0 = _mm_sub_ss(_xmm0, _xmm1);
	 _xmm0 = _mm_mul_ss(_xmm0, _xmm7);
	 _mm_store_ss(&outRotate->v.w, _xmm0);
 }

 return TRUE;
}
///////////////////////////////////////////////////////////////////////////////	
BOOL c_MATRIX::Decompose(c_VECTOR *outScale, 
					     c_VECTOR *outRotate,
					     c_VECTOR *outTranslate) const
{
 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6, _xmm7;
 float AlReg[4] = { 0 };

 _xmm0 = _mm_loadu_ps(&m.v[ 0]);
 _xmm1 = _mm_loadu_ps(&m.v[ 4]);
 _xmm2 = _mm_loadu_ps(&m.v[ 8]);

 *outTranslate = *((c_VECTOR*)&m.v[12]);

 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm1, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm5 = _mm_shuffle_ps(_xmm5, _xmm2, _MM_SHUFFLE(0, 0, 2, 0));

 _xmm4 = _mm_mul_ps(_xmm5, _xmm5);
 _xmm5 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm6 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(2, 2, 2, 2));

 _xmm4 = _mm_add_ss(_xmm4, _xmm5);
 _xmm4 = _mm_add_ss(_xmm4, _xmm6);

 _xmm4 = _mm_rsqrt_ss(_xmm4);

 _mm_store_ss(&AlReg[0], _xmm4);

 _xmm4 = _mm_rcp_ss(_xmm4);

 // _xmm4.x = scaleX

 _mm_store_ss(&outScale->v.x, _xmm4);

 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm1, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm5 = _mm_shuffle_ps(_xmm5, _xmm2, _MM_SHUFFLE(1, 1, 2, 0));

 _xmm4 = _mm_mul_ps(_xmm5, _xmm5);
 _xmm5 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm6 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(2, 2, 2, 2));

 _xmm4 = _mm_add_ss(_xmm4, _xmm5);
 _xmm4 = _mm_add_ss(_xmm4, _xmm6);

 _xmm4 = _mm_rsqrt_ss(_xmm4);

 _mm_store_ss(&AlReg[1], _xmm4);

 _xmm4 = _mm_rcp_ss(_xmm4);

 // _xmm4.x = scaleY

 _mm_store_ss(&outScale->v.y, _xmm4);

 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm5 = _mm_shuffle_ps(_xmm5, _xmm2, _MM_SHUFFLE(2, 2, 2, 0));

 _xmm4 = _mm_mul_ps(_xmm5, _xmm5);
 _xmm5 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm6 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(2, 2, 2, 2));

 _xmm4 = _mm_add_ss(_xmm4, _xmm5);
 _xmm4 = _mm_add_ss(_xmm4, _xmm6);

 _xmm4 = _mm_rsqrt_ss(_xmm4);

 _mm_store_ss(&AlReg[2], _xmm4);

 _xmm4 = _mm_rcp_ss(_xmm4);
 _xmm7 = _mm_loadu_ps(AlReg);

 // _xmm4.x = scaleZ

 _mm_store_ss(&outScale->v.z, _xmm4);

 _xmm0 = _mm_mul_ps(_xmm0, _xmm7);
 _xmm1 = _mm_mul_ps(_xmm1, _xmm7);
 _xmm2 = _mm_mul_ps(_xmm2, _xmm7);

 //
 // _xmm[0-2] Normalized rotation matrix that need to be converted to QUAT
 //
 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0)); // xmm0.xxxx -> xmm4
 _xmm5 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1)); // xmm1.yyyy -> xmm5
 _xmm6 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 2, 2, 2)); // xmm2.zzzz -> xmm6

 if (_mm_comigt_ss(_xmm4, _xmm5) && _mm_comigt_ss(_xmm4, _xmm6))
 {
	 _xmm7 = _mm_add_ss(_xmm5,    _xmm6);
	 _xmm7 = _mm_sub_ss(_xmm4,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 *outRotate = c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
		 return TRUE;
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7); 
	 _xmm7 = _mm_rcp_ss(_xmm7); 
	 _mm_store_ss(&outRotate->v.x, _xmm3);

	 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm4 = _mm_add_ss(_xmm4, _xmm1);
	 _xmm4 = _mm_mul_ss(_xmm4, _xmm7);
	 _mm_store_ss(&outRotate->v.y, _xmm4);

	 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm2);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);	
	 _mm_store_ss(&outRotate->v.z, _xmm5);

	 _xmm6 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm6 = _mm_sub_ss(_xmm6, _xmm2);
	 _xmm6 = _mm_mul_ss(_xmm6, _xmm7);
	 _mm_store_ss(&outRotate->v.w, _xmm6);
 } 
 else 
 if (_mm_comigt_ss(_xmm5, _xmm6))
 {
	 _xmm7 = _mm_add_ss(_xmm4,    _xmm6);
	 _xmm7 = _mm_sub_ss(_xmm5,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 *outRotate = c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
		 return TRUE;
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7);
	 _xmm7 = _mm_rcp_ss(_xmm7);             
	 _mm_store_ss(&outRotate->v.y, _xmm3);

	 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm1);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);
	 _mm_store_ss(&outRotate->v.x, _xmm5);

	 _xmm6 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm3 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm6 = _mm_add_ss(_xmm6, _xmm3);
	 _xmm6 = _mm_mul_ss(_xmm6, _xmm7);
	 _mm_store_ss(&outRotate->v.z, _xmm6);

	 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm2 = _mm_sub_ss(_xmm2, _xmm0);
	 _xmm2 = _mm_mul_ss(_xmm2, _xmm7);
	 _mm_store_ss(&outRotate->v.w, _xmm2);
 } 
 else
 {
	 _xmm7 = _mm_add_ss(_xmm4,    _xmm5);
	 _xmm7 = _mm_sub_ss(_xmm6,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 *outRotate = c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
		 return TRUE;
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7);
	 _xmm7 = _mm_rcp_ss(_xmm7); 
	 _mm_store_ss(&outRotate->v.z, _xmm3);

	 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm4 = _mm_add_ss(_xmm4, _xmm2);
	 _xmm4 = _mm_mul_ss(_xmm4, _xmm7);
	 _mm_store_ss(&outRotate->v.x, _xmm4);

	 _xmm5 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm6 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm6);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);
	 _mm_store_ss(&outRotate->v.y, _xmm5);

	 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm0 = _mm_sub_ss(_xmm0, _xmm1);
	 _xmm0 = _mm_mul_ss(_xmm0, _xmm7);
	 _mm_store_ss(&outRotate->v.w, _xmm0);
 }

 return TRUE;
}
///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::Compose(const c_VECTOR *inScale,
					   const c_VECTOR *inRotate,
					   const c_VECTOR *inTranslate)
{
 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5;

 _xmm0 = _mm_loadu_ps  (inRotate->v.v);

 _xmm1 = _mm_mul_ps    (_xmm0, _xmm0); 

 _xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm2 = _mm_mul_ps    (_xmm2, _xmm0);

 _xmm3 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 1, 1));
 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 2));
 _xmm3 = _mm_mul_ps    (_xmm3, _xmm4);

 _xmm0 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(1, 2, 1, 2));
 _xmm0 = _mm_mul_ps    (_xmm0, __xmmSGNM);
 _xmm4 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 1, 2, 1));
 _xmm4 = _mm_add_ps    (_xmm4, _xmm0);      // y0, z0, x1, x2

 _xmm0 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 1, 0));
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 2, 2));
 _xmm0 = _mm_add_ps    (_xmm0, _xmm1);      // y1, x0, z2

 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(3, 3, 3, 3));
 _xmm2 = _mm_mul_ps    (_xmm2, __xmmSGNM);
 _xmm3 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm3 = _mm_add_ps    (_xmm3, _xmm2);      // z1, y2

 _xmm0 = _mm_add_ps    (_xmm0, _xmm0);      // y1, x0, z2     | * 2
 _xmm4 = _mm_add_ps    (_xmm4, _xmm4);      // y0, z0, x1, x2 | * 2
 _xmm3 = _mm_add_ps    (_xmm3, _xmm3);      // z1, y2		  | * 2
 _xmm0 = _mm_sub_ps    (__xmmONE, _xmm0);   // y1, x0, z2     | 1 - 

 //
 // Swizzle output
 //
 _xmm5 = _mm_loadu_ps(inScale->v.v);
 _xmm5 = _mm_mul_ps(_xmm5, __xmmMASK);

 _xmm1 = _mm_movelh_ps (_xmm0, _xmm4);
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 3, 2, 1));
 _xmm1 = _mm_mul_ps    (_xmm1, _xmm5);

 _xmm3 = _mm_movelh_ps (_xmm3, _xmm0);
 _xmm2 = _mm_shuffle_ps(_xmm3, _xmm4, _MM_SHUFFLE(2, 2, 2, 0));
 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(0, 0, 1, 2));
 _xmm2 = _mm_mul_ps    (_xmm2, _xmm5);

 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm3, _MM_SHUFFLE(1, 1, 3, 3));
 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm4 = _mm_move_ss   (_xmm4, _xmm0);
 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(0, 0, 2, 1));
 _xmm4 = _mm_mul_ps    (_xmm4, _xmm5);

 _mm_storeu_ps		   (&m.v[0],  _xmm1);
 _mm_storeu_ps		   (&m.v[4],  _xmm2);
 _mm_storeu_ps		   (&m.v[8],  _xmm4);

 _xmm4 = _mm_loadu_ps(inTranslate->v.v);
 _xmm4 = _mm_mul_ps(_xmm4, __xmmMASK);
 _xmm4 = _mm_add_ps(_xmm4, __xmmTSSK);
 _mm_storeu_ps	   (&m.v[12], _xmm4);

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_MATRIX::FromQuat(const c_VECTOR &Quat)
{
 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4;

 _xmm0 = _mm_loadu_ps  (Quat.v.v);          // Load quaternion

 _xmm1 = _mm_mul_ps    (_xmm0, _xmm0); 

 _xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm2 = _mm_mul_ps    (_xmm2, _xmm0);

 _xmm3 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 1, 1));
 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 2));
 _xmm3 = _mm_mul_ps    (_xmm3, _xmm4);

 _xmm0 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(1, 2, 1, 2));
 _xmm0 = _mm_mul_ps    (_xmm0, __xmmSGNM);
 _xmm4 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 1, 2, 1));
 _xmm4 = _mm_add_ps    (_xmm4, _xmm0);      // y0, z0, x1, x2

 _xmm0 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 1, 0));
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 2, 2));
 _xmm0 = _mm_add_ps    (_xmm0, _xmm1);      // y1, x0, z2

 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(3, 3, 3, 3));
 _xmm2 = _mm_mul_ps    (_xmm2, __xmmSGNM);
 _xmm3 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm3 = _mm_add_ps    (_xmm3, _xmm2);      // z1, y2

 _xmm0 = _mm_add_ps    (_xmm0, _xmm0);      // y1, x0, z2     | * 2
 _xmm4 = _mm_add_ps    (_xmm4, _xmm4);      // y0, z0, x1, x2 | * 2
 _xmm3 = _mm_add_ps    (_xmm3, _xmm3);      // z1, y2		  | * 2
 _xmm0 = _mm_sub_ps    (__xmmONE, _xmm0);   // y1, x0, z2     | 1 - 

 //
 // Swizzle output
 //
 _xmm1 = _mm_movelh_ps (_xmm0, _xmm4);
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 3, 2, 1));
 _xmm1 = _mm_mul_ps    (_xmm1, __xmmMASK);

 _xmm3 = _mm_movelh_ps (_xmm3, _xmm0);
 _xmm2 = _mm_shuffle_ps(_xmm3, _xmm4, _MM_SHUFFLE(2, 2, 2, 0));
 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(0, 0, 1, 2));
 _xmm2 = _mm_mul_ps    (_xmm2, __xmmMASK);

 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm3, _MM_SHUFFLE(1, 1, 3, 3));
 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm4 = _mm_move_ss   (_xmm4, _xmm0);
 _xmm4 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(0, 0, 2, 1));
 _xmm4 = _mm_mul_ps    (_xmm4, __xmmMASK);

 _mm_storeu_ps		   (&m.v[0],  _xmm1);
 _mm_storeu_ps		   (&m.v[4],  _xmm2);
 _mm_storeu_ps		   (&m.v[8],  _xmm4);
 _mm_storeu_ps		   (&m.v[12], __xmmTSSK);

 return;
}
///////////////////////////////////////////////////////////////////////////////	
c_VECTOR c_MATRIX::ToQuat() const
{
 c_VECTOR output;

 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6, _xmm7;

 _xmm0 = _mm_loadu_ps(&m.v[ 0]);
 _xmm1 = _mm_loadu_ps(&m.v[ 4]);
 _xmm2 = _mm_loadu_ps(&m.v[ 8]);

 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0)); // xmm0.xxxx -> xmm4
 _xmm5 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1)); // xmm1.yyyy -> xmm5
 _xmm6 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(2, 2, 2, 2)); // xmm2.zzzz -> xmm6

 if (_mm_comigt_ss(_xmm4, _xmm5) && _mm_comigt_ss(_xmm4, _xmm6))
 {
	 _xmm7 = _mm_add_ss(_xmm5,    _xmm6);
	 _xmm7 = _mm_sub_ss(_xmm4,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 return c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7); 
	 _xmm7 = _mm_rcp_ss(_xmm7); 
	 _mm_store_ss(&output.v.x, _xmm3);

	 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm4 = _mm_add_ss(_xmm4, _xmm1);
	 _xmm4 = _mm_mul_ss(_xmm4, _xmm7);
	 _mm_store_ss(&output.v.y, _xmm4);

	 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm2);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);	
	 _mm_store_ss(&output.v.z, _xmm5);

	 _xmm6 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm6 = _mm_sub_ss(_xmm6, _xmm2);
	 _xmm6 = _mm_mul_ss(_xmm6, _xmm7);
	 _mm_store_ss(&output.v.w, _xmm6);
 } 
 else 
 if (_mm_comigt_ss(_xmm5, _xmm6))
 {
	 _xmm7 = _mm_add_ss(_xmm4,    _xmm6);
	 _xmm7 = _mm_sub_ss(_xmm5,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 return c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7);
	 _xmm7 = _mm_rcp_ss(_xmm7);             
	 _mm_store_ss(&output.v.y, _xmm3);

	 _xmm5 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm1);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);
	 _mm_store_ss(&output.v.x, _xmm5);

	 _xmm6 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm3 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm6 = _mm_add_ss(_xmm6, _xmm3);
	 _xmm6 = _mm_mul_ss(_xmm6, _xmm7);
	 _mm_store_ss(&output.v.z, _xmm6);

	 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm2 = _mm_sub_ss(_xmm2, _xmm0);
	 _xmm2 = _mm_mul_ss(_xmm2, _xmm7);
	 _mm_store_ss(&output.v.w, _xmm2);
 } 
 else
 {
	 _xmm7 = _mm_add_ss(_xmm4,    _xmm5);
	 _xmm7 = _mm_sub_ss(_xmm6,    _xmm7);
	 _xmm7 = _mm_add_ss(__xmmONE, _xmm7);

	 if (_mm_comilt_ss(_xmm7, __xmmEPS))
	 {
		 return c_VECTOR(0.0f, 0.0f, 0.0f, 1.0f);
	 }

	 _xmm7 = _mm_sqrt_ss(_xmm7);
	 _xmm7 = _mm_add_ss(_xmm7, _xmm7);
	 _xmm3 = _mm_mul_ss(__xmmQUAD, _xmm7);
	 _xmm7 = _mm_rcp_ss(_xmm7); 
	 _mm_store_ss(&output.v.z, _xmm3);

	 _xmm4 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm4 = _mm_add_ss(_xmm4, _xmm2);
	 _xmm4 = _mm_mul_ss(_xmm4, _xmm7);
	 _mm_store_ss(&output.v.x, _xmm4);

	 _xmm5 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2));
	 _xmm6 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm5 = _mm_add_ss(_xmm5, _xmm6);
	 _xmm5 = _mm_mul_ss(_xmm5, _xmm7);
	 _mm_store_ss(&output.v.y, _xmm5);

	 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
	 _xmm0 = _mm_sub_ss(_xmm0, _xmm1);
	 _xmm0 = _mm_mul_ss(_xmm0, _xmm7);
	 _mm_store_ss(&output.v.w, _xmm0);
 }

 return output;
}
///////////////////////////////////////////////////////////////////////////////	
BOOL c_MATRIX::Inverse()
{
 float *src = (float *)this;
 __m128 minor0, minor1, minor2, minor3;
 __m128 row0, row1, row2, row3;
 __m128 det, tmp1;

 tmp1   = _mm_loadh_pi(_mm_loadl_pi(tmp1, (__m64*)(src)), (__m64*)(src+ 4));
 row1   = _mm_loadh_pi(_mm_loadl_pi(row1, (__m64*)(src+8)), (__m64*)(src+12));
 row0   = _mm_shuffle_ps(tmp1, row1, 0x88);
 row1   = _mm_shuffle_ps(row1, tmp1, 0xDD);
 tmp1   = _mm_loadh_pi(_mm_loadl_pi(tmp1, (__m64*)(src+ 2)), (__m64*)(src+ 6));
 row3   = _mm_loadh_pi(_mm_loadl_pi(row3, (__m64*)(src+10)), (__m64*)(src+14));
 row2   = _mm_shuffle_ps(tmp1, row3, 0x88);
 row3   = _mm_shuffle_ps(row3, tmp1, 0xDD);

 tmp1   = _mm_mul_ps(row2, row3);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0xB1);
 minor0 = _mm_mul_ps(row1, tmp1);
 minor1 = _mm_mul_ps(row0, tmp1);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0x4E);
 minor0 = _mm_sub_ps(_mm_mul_ps(row1, tmp1), minor0);
 minor1 = _mm_sub_ps(_mm_mul_ps(row0, tmp1), minor1);
 minor1 = _mm_shuffle_ps(minor1, minor1, 0x4E);

 tmp1   = _mm_mul_ps(row1, row2);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0xB1);
 minor0 = _mm_add_ps(_mm_mul_ps(row3, tmp1), minor0);
 minor3 = _mm_mul_ps(row0, tmp1);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0x4E);
 minor0 = _mm_sub_ps(minor0, _mm_mul_ps(row3, tmp1));
 minor3 = _mm_sub_ps(_mm_mul_ps(row0, tmp1), minor3);
 minor3 = _mm_shuffle_ps(minor3, minor3, 0x4E);

 tmp1   = _mm_mul_ps(_mm_shuffle_ps(row1, row1, 0x4E), row3);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0xB1);
 row2   = _mm_shuffle_ps(row2, row2, 0x4E);
 minor0 = _mm_add_ps(_mm_mul_ps(row2, tmp1), minor0);
 minor2 = _mm_mul_ps(row0, tmp1);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0x4E);
 minor0 = _mm_sub_ps(minor0, _mm_mul_ps(row2, tmp1));
 minor2 = _mm_sub_ps(_mm_mul_ps(row0, tmp1), minor2);
 minor2 = _mm_shuffle_ps(minor2, minor2, 0x4E);

 tmp1   = _mm_mul_ps(row0, row1);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0xB1);
 minor2 = _mm_add_ps(_mm_mul_ps(row3, tmp1), minor2);
 minor3 = _mm_sub_ps(_mm_mul_ps(row2, tmp1), minor3);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0x4E);
 minor2 = _mm_sub_ps(_mm_mul_ps(row3, tmp1), minor2);
 minor3 = _mm_sub_ps(minor3, _mm_mul_ps(row2, tmp1)); 

 tmp1   = _mm_mul_ps(row0, row3);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0xB1);
 minor1 = _mm_sub_ps(minor1, _mm_mul_ps(row2, tmp1));
 minor2 = _mm_add_ps(_mm_mul_ps(row1, tmp1), minor2);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0x4E);
 minor1 = _mm_add_ps(_mm_mul_ps(row2, tmp1), minor1);
 minor2 = _mm_sub_ps(minor2, _mm_mul_ps(row1, tmp1));

 tmp1   = _mm_mul_ps(row0, row2);
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0xB1);
 minor1 = _mm_add_ps(_mm_mul_ps(row3, tmp1), minor1);
 minor3 = _mm_sub_ps(minor3, _mm_mul_ps(row1, tmp1));
 tmp1   = _mm_shuffle_ps(tmp1, tmp1, 0x4E);
 minor1 = _mm_sub_ps(minor1, _mm_mul_ps(row3, tmp1));
 minor3 = _mm_add_ps(_mm_mul_ps(row1, tmp1), minor3);

 det    = _mm_mul_ps(row0, minor0);
 det    = _mm_add_ps(_mm_shuffle_ps(det, det, 0x4E), det);
 det    = _mm_add_ss(_mm_shuffle_ps(det, det, 0xB1), det);
 tmp1   = _mm_rcp_ss(det);
 det    = _mm_sub_ss(_mm_add_ss(tmp1, tmp1), _mm_mul_ss(det, _mm_mul_ss(tmp1, tmp1)));
 det    = _mm_shuffle_ps(det, det, 0x00);
 minor0 = _mm_mul_ps(det, minor0);

 _mm_storel_pi((__m64*)(src), minor0);
 _mm_storeh_pi((__m64*)(src+2), minor0);
 minor1 = _mm_mul_ps(det, minor1);
 _mm_storel_pi((__m64*)(src+4), minor1);
 _mm_storeh_pi((__m64*)(src+6), minor1);
 minor2 = _mm_mul_ps(det, minor2);
 _mm_storel_pi((__m64*)(src+ 8), minor2);
 _mm_storeh_pi((__m64*)(src+10), minor2);
 minor3 = _mm_mul_ps(det, minor3);
 _mm_storel_pi((__m64*)(src+12), minor3);
 _mm_storeh_pi((__m64*)(src+14), minor3);

 return TRUE;
}
///////////////////////////////////////////////////////////////////////////////	
