#ifndef __MATRIX_H_INCLUDED__
#define __MATRIX_H_INCLUDED__

///////////////////////////////////////////////////////////////////////////////
typedef struct _c_MATRIX_STR
{
 union {
        struct  {				//  v--- unused
                  float xx, xy, xz, xw;
                  float yx, yy, yz, yw;
                  float zx, zy, zz, zw;
                  float wx, wy, wz, ww; // <-- translation
                };
        struct  {
                  float  _00, _01, _02, _03;
                  float  _10, _11, _12, _13;
                  float  _20, _21, _22, _23;
                  float  _30, _31, _32, _33;
                };
        struct  {
                  float v[16];
                };
       };

}c_MATRIX_STR;

__forceinline __m128 __FAST_MULLVEC3(__m128 *mr, __m128 vec)
{
 __m128 _xmm0, _xmm1, _xmm2;

 _xmm0 = _mm_shuffle_ps(vec, vec, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm1 = _mm_shuffle_ps(vec, vec, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm2 = _mm_shuffle_ps(vec, vec, _MM_SHUFFLE(2, 2, 2, 2));

 _xmm0 = _mm_mul_ps(_xmm0, mr[0]);
 _xmm1 = _mm_mul_ps(_xmm1, mr[1]);
 _xmm2 = _mm_mul_ps(_xmm2, mr[2]);

 _xmm0 = _mm_add_ps(_xmm0, _xmm1);
 _xmm0 = _mm_add_ps(_xmm0, _xmm2);
 _xmm0 = _mm_add_ps(_xmm0, mr[3]);

 return _xmm0;
}

__forceinline __m128 __FAST_MULLVEC4(__m128 *mr, __m128 vec)
{
 __m128 _xmm0, _xmm1, _xmm2, _xmm3;

 _xmm0 = _mm_shuffle_ps(vec, vec, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm1 = _mm_shuffle_ps(vec, vec, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm2 = _mm_shuffle_ps(vec, vec, _MM_SHUFFLE(2, 2, 2, 2));
 _xmm3 = _mm_shuffle_ps(vec, vec, _MM_SHUFFLE(3, 3, 3, 3));

 _xmm0 = _mm_mul_ps(_xmm0, mr[0]);
 _xmm1 = _mm_mul_ps(_xmm1, mr[1]);
 _xmm2 = _mm_mul_ps(_xmm2, mr[2]);
 _xmm3 = _mm_mul_ps(_xmm3, mr[3]);

 _xmm0 = _mm_add_ps(_xmm0, _xmm1);
 _xmm0 = _mm_add_ps(_xmm0, _xmm2);
 _xmm0 = _mm_add_ps(_xmm0, _xmm3);

 return _xmm0;
}

class c_MATRIX
{
	private:

		c_MATRIX(int _duumy)
		{
			// Empty constructor
		}

		__forceinline void Mull(const c_MATRIX *m1, const c_MATRIX *m2)
		{
			__m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6, _xmm7;
			__m128 _tm0, _tm1, _tm2, _tm3;

			_tm0  = _mm_loadu_ps  (&m2->m.v[ 0]);
			_tm1  = _mm_loadu_ps  (&m2->m.v[ 4]);
			_tm2  = _mm_loadu_ps  (&m2->m.v[ 8]);
			_tm3  = _mm_loadu_ps  (&m2->m.v[12]);

			_xmm4 = _mm_loadu_ps  (&m1->m.v[ 0]);
			_xmm5 = _mm_loadu_ps  (&m1->m.v[ 4]);
			_xmm6 = _mm_loadu_ps  (&m1->m.v[ 8]);
			_xmm7 = _mm_loadu_ps  (&m1->m.v[12]);

			_xmm0 = _mm_shuffle_ps(_tm0,  _tm0, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm1 = _mm_shuffle_ps(_tm0,  _tm0, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm2 = _mm_shuffle_ps(_tm0,  _tm0, _MM_SHUFFLE(2, 2, 2, 2));
			_xmm3 = _mm_shuffle_ps(_tm0,  _tm0, _MM_SHUFFLE(3, 3, 3, 3));
			_xmm0 = _mm_mul_ps	  (_xmm0, _xmm4);
			_xmm1 = _mm_mul_ps	  (_xmm1, _xmm5);
			_xmm2 = _mm_mul_ps	  (_xmm2, _xmm6);
			_xmm3 = _mm_mul_ps	  (_xmm3, _xmm7);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm1);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm2);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm3);
			_mm_storeu_ps		  (&m.v[ 0], _xmm0);

			_xmm0 = _mm_shuffle_ps(_tm1,  _tm1, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm1 = _mm_shuffle_ps(_tm1,  _tm1, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm2 = _mm_shuffle_ps(_tm1,  _tm1, _MM_SHUFFLE(2, 2, 2, 2));
			_xmm3 = _mm_shuffle_ps(_tm1,  _tm1, _MM_SHUFFLE(3, 3, 3, 3));
			_xmm0 = _mm_mul_ps	  (_xmm0, _xmm4);
			_xmm1 = _mm_mul_ps	  (_xmm1, _xmm5);
			_xmm2 = _mm_mul_ps	  (_xmm2, _xmm6);
			_xmm3 = _mm_mul_ps	  (_xmm3, _xmm7);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm1);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm2);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm3);
			_mm_storeu_ps		  (&m.v[ 4], _xmm0);

			_xmm0 = _mm_shuffle_ps(_tm2,  _tm2, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm1 = _mm_shuffle_ps(_tm2,  _tm2, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm2 = _mm_shuffle_ps(_tm2,  _tm2, _MM_SHUFFLE(2, 2, 2, 2));
			_xmm3 = _mm_shuffle_ps(_tm2,  _tm2, _MM_SHUFFLE(3, 3, 3, 3));
			_xmm0 = _mm_mul_ps	  (_xmm0, _xmm4);
			_xmm1 = _mm_mul_ps	  (_xmm1, _xmm5);
			_xmm2 = _mm_mul_ps	  (_xmm2, _xmm6);
			_xmm3 = _mm_mul_ps	  (_xmm3, _xmm7);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm1);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm2);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm3);
			_mm_storeu_ps		  (&m.v[ 8], _xmm0);

			_xmm0 = _mm_shuffle_ps(_tm3,  _tm3, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm1 = _mm_shuffle_ps(_tm3,  _tm3, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm2 = _mm_shuffle_ps(_tm3,  _tm3, _MM_SHUFFLE(2, 2, 2, 2));
			_xmm3 = _mm_shuffle_ps(_tm3,  _tm3, _MM_SHUFFLE(3, 3, 3, 3));
			_xmm0 = _mm_mul_ps	  (_xmm0, _xmm4);
			_xmm1 = _mm_mul_ps	  (_xmm1, _xmm5);
			_xmm2 = _mm_mul_ps	  (_xmm2, _xmm6);
			_xmm3 = _mm_mul_ps	  (_xmm3, _xmm7);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm1);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm2);
			_xmm0 = _mm_add_ps	  (_xmm0, _xmm3);
			_mm_storeu_ps		  (&m.v[12], _xmm0);

			return;
		}

    public:

        c_MATRIX_STR m;

		c_MATRIX(float _00, float _01, float _02, float _03,
                 float _10, float _11, float _12, float _13,
                 float _20, float _21, float _22, float _23,
                 float _30, float _31, float _32, float _33)
		{
			m.v[ 0] = _00;
			m.v[ 1] = _01;
			m.v[ 2] = _02;
			m.v[ 3] = _03;

			m.v[ 4] = _10;
			m.v[ 5] = _11;
			m.v[ 6] = _12;
			m.v[ 7] = _13;

			m.v[ 8] = _20;
			m.v[ 9] = _21;
			m.v[10] = _22;
			m.v[11] = _23;

			m.v[12] = _30;
			m.v[13] = _31;
			m.v[14] = _32;
			m.v[15] = _33;

			return;
		}

		c_MATRIX()
		{
			Identity();
		}

		void Identity()
		{
			for (int i = 0; i < 16; i++)
			     m.v[i] = 0.0f;
 
			m.xx = m.yy = m.zz = m.ww = 1.0f;
			return;
		}

		BOOL IsIdentity() const
		{
			if (fabsf(m.v[ 0] - 1.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[ 1] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[ 2] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[ 3] - 0.0f) > MATH_EPSILON) return FALSE;

			if (fabsf(m.v[ 4] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[ 5] - 1.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[ 6] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[ 7] - 0.0f) > MATH_EPSILON) return FALSE;

			if (fabsf(m.v[ 8] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[ 9] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[10] - 1.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[11] - 0.0f) > MATH_EPSILON) return FALSE;

			if (fabsf(m.v[12] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[13] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[14] - 0.0f) > MATH_EPSILON) return FALSE;
			if (fabsf(m.v[15] - 1.0f) > MATH_EPSILON) return FALSE;

			return TRUE;
		}

		void Transpose()
		{
			#define __XCH(a, b)  { tmp = a; a = b; b = tmp; }
			float tmp;
 
			__XCH(m.v[ 1], m.v[ 4]);
			__XCH(m.v[ 2], m.v[ 8]);
			__XCH(m.v[ 3], m.v[12]);
			__XCH(m.v[ 6], m.v[ 9]);
			__XCH(m.v[ 7], m.v[13]);
			__XCH(m.v[11], m.v[14]);

			#undef __XCH
			return;
		}

		void MullVec3(const c_VECTOR *a, c_VECTOR *_out) const
		{
			__m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6;

			_xmm0 = _mm_loadu_ps(a->v.v);

			_xmm3 = _mm_loadu_ps(&m.v[0]);
			_xmm4 = _mm_loadu_ps(&m.v[4]);
			_xmm5 = _mm_loadu_ps(&m.v[8]);
			_xmm6 = _mm_loadu_ps(&m.v[12]);

			_xmm1 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));

			_xmm1 = _mm_mul_ps(_xmm1, _xmm3);
			_xmm2 = _mm_mul_ps(_xmm2, _xmm4);
			_xmm0 = _mm_mul_ps(_xmm0, _xmm5);

			_xmm1 = _mm_add_ps(_xmm1, _xmm2);
			_xmm1 = _mm_add_ps(_xmm1, _xmm0);
			_xmm1 = _mm_add_ps(_xmm1, _xmm6);

			_mm_storeu_ps(_out->v.v, _xmm1);
		}

		void MullVec4( const c_VECTOR *a, c_VECTOR *_out ) const
		{
			__m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6, _xmm7;

			_xmm0 = _mm_loadu_ps(a->v.v);

			_xmm3 = _mm_loadu_ps(&m.v[0]);
			_xmm4 = _mm_loadu_ps(&m.v[4]);
			_xmm5 = _mm_loadu_ps(&m.v[8]);
			_xmm6 = _mm_loadu_ps(&m.v[12]);

			_xmm1 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm7 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));
			_xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 3));

			_xmm1 = _mm_mul_ps(_xmm1, _xmm3);
			_xmm2 = _mm_mul_ps(_xmm2, _xmm4);
			_xmm7 = _mm_mul_ps(_xmm7, _xmm5);
			_xmm0 = _mm_mul_ps(_xmm0, _xmm6);

			_xmm1 = _mm_add_ps(_xmm1, _xmm2);
			_xmm1 = _mm_add_ps(_xmm1, _xmm7);
			_xmm1 = _mm_add_ps(_xmm1, _xmm0);

			_mm_storeu_ps(_out->v.v, _xmm1);
			return;
		}

		void MullNorm3( const c_VECTOR *a, c_VECTOR *_out ) const
		{
			__m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5;

			_xmm0 = _mm_loadu_ps(a->v.v);

			_xmm3 = _mm_loadu_ps(&m.v[0]);
			_xmm4 = _mm_loadu_ps(&m.v[4]);
			_xmm5 = _mm_loadu_ps(&m.v[8]);

			_xmm1 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 2, 2, 2));

			_xmm1 = _mm_mul_ps(_xmm1, _xmm3);
			_xmm2 = _mm_mul_ps(_xmm2, _xmm4);
			_xmm0 = _mm_mul_ps(_xmm0, _xmm5);

			_xmm1 = _mm_add_ps(_xmm1, _xmm2);
			_xmm1 = _mm_add_ps(_xmm1, _xmm0);

			_mm_storeu_ps(_out->v.v, _xmm1);

		    return;
		}

		c_MATRIX operator * (const c_MATRIX &oth) const
		{
			c_MATRIX tmp(0);

			tmp.Mull((c_MATRIX *)this, &oth);

			return tmp;
		}

		void operator *= (const c_MATRIX &oth)
		{
			c_MATRIX tmp(0);

			tmp.Mull((c_MATRIX *)this, &oth);

			*this = tmp;
		}

        c_MATRIX operator + (const c_MATRIX &oth) const
		{
			c_MATRIX n(0);

			_mm_storeu_ps(&n.m.v[0],  _mm_add_ps(_mm_loadu_ps(&m.v[0]),  _mm_loadu_ps(&oth.m.v[0])));
			_mm_storeu_ps(&n.m.v[4],  _mm_add_ps(_mm_loadu_ps(&m.v[4]),  _mm_loadu_ps(&oth.m.v[4])));
			_mm_storeu_ps(&n.m.v[8],  _mm_add_ps(_mm_loadu_ps(&m.v[8]),  _mm_loadu_ps(&oth.m.v[8])));
			_mm_storeu_ps(&n.m.v[12], _mm_add_ps(_mm_loadu_ps(&m.v[12]), _mm_loadu_ps(&oth.m.v[12])));

			return n;
		}

		c_MATRIX operator += (const c_MATRIX &oth)
		{
			_mm_storeu_ps(&m.v[0],  _mm_add_ps(_mm_loadu_ps(&m.v[0]),  _mm_loadu_ps(&oth.m.v[0])));
			_mm_storeu_ps(&m.v[4],  _mm_add_ps(_mm_loadu_ps(&m.v[4]),  _mm_loadu_ps(&oth.m.v[4])));
			_mm_storeu_ps(&m.v[8],  _mm_add_ps(_mm_loadu_ps(&m.v[8]),  _mm_loadu_ps(&oth.m.v[8])));
			_mm_storeu_ps(&m.v[12], _mm_add_ps(_mm_loadu_ps(&m.v[12]), _mm_loadu_ps(&oth.m.v[12])));

			return *this;
		}

		c_MATRIX operator * (const float &oth) const
		{
			c_MATRIX tmp(0);

			__m128 _xmm0 = _mm_set_ps1(oth);

			_mm_storeu_ps(&tmp.m.v[0],  _mm_mul_ps(_mm_loadu_ps(&m.v[0]),  _xmm0));
			_mm_storeu_ps(&tmp.m.v[4],  _mm_mul_ps(_mm_loadu_ps(&m.v[4]),  _xmm0));
			_mm_storeu_ps(&tmp.m.v[8],  _mm_mul_ps(_mm_loadu_ps(&m.v[8]),  _xmm0));
			_mm_storeu_ps(&tmp.m.v[12], _mm_mul_ps(_mm_loadu_ps(&m.v[12]), _xmm0));

			return tmp;
		}

		c_VECTOR operator * (const c_VECTOR &vec) const
		{
			__m128 tmp[4];
				   tmp[0] = _mm_loadu_ps(&m.v[0]);
				   tmp[1] = _mm_loadu_ps(&m.v[4]);
				   tmp[2] = _mm_loadu_ps(&m.v[8]);
				   tmp[3] = _mm_loadu_ps(&m.v[12]);

			return __FAST_MULLVEC4(tmp, _mm_loadu_ps(vec.v.v));
		}

		void DebugCompare(const char *message, const c_MATRIX &other) const
		{
			mxLogger("Comparing: [%s]", message);
			for (int i = 0; i < 16; i++)
			{
				if (fabsf(m.v[i] - other.m.v[i]) > MATH_EPSILON)
					mxLogger("%i : [%f] != [%f]", i, m.v[i], other.m.v[i]);
			}
			mxLogger( "Done" );
		}

		void DebugPrint(const char *message) const
		{
			mxLogger("Debug printout: [%s]", message);
			for (int i = 0; i < 4; i++)
			     mxLogger("[%6.4f] [%6.4f] [%6.4f] [%6.4f]", m.v[(i*4)+0], m.v[(i*4)+1], m.v[(i*4)+2], m.v[(i*4)+3]);
			mxLogger("Done");
			return;
		}

		void Zero()
		{
			memset(m.v, 0, sizeof(c_MATRIX));
		}

		//
		// Make perspective projection matrix
		//
		void MakeProjectionMatrix(float FOV, float Aspect, float NearPlane, float FarPlane);

		//
		// Make 'look at' matrix
		//
		void MakeLookAtMatrix(float eyex,    float eyey,    float eyez,
		                      float centerx, float centery, float centerz,
					          float upx,     float upy,     float upz);

		//
		// Make orthographic projection matrix
		//
        void MakeOrthoMatrix(float l, float r, float b, float t, float n, float f);

		//
		// Convert from quat
		//
		void FromQuat(const c_VECTOR &quat);

		//
		// Rotations
		//
		void MakeRotation(float angle, float x, float y, float z)
		{
		   float mag, s, c;
		   float xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;

		   _fpusincos(angle, s, c);

		   mag = sqrtf(x*x + y*y + z*z);

		   if (mag < MATH_EPSILON)
		   {
			  this->Identity();
		      return;
		   }

		   x /= mag;
		   y /= mag;
		   z /= mag;

		   #define M(row,col)  m.v[(col<<2)+row]

		   xx = x * x;
		   yy = y * y;
		   zz = z * z;
		   xy = x * y;
		   yz = y * z;
		   zx = z * x;
		   xs = x * s;
		   ys = y * s;
		   zs = z * s;
		   one_c = 1.0f - c;

		   M(0,0) = (one_c * xx) + c;
		   M(0,1) = (one_c * xy) - zs;
		   M(0,2) = (one_c * zx) + ys;
		   M(0,3) = 0.0f;

		   M(1,0) = (one_c * xy) + zs;
		   M(1,1) = (one_c * yy) + c;
		   M(1,2) = (one_c * yz) - xs;
		   M(1,3) = 0.0f;

		   M(2,0) = (one_c * zx) - ys;
		   M(2,1) = (one_c * yz) + xs;
		   M(2,2) = (one_c * zz) + c;
		   M(2,3) = 0.0f;

		   M(3,0) = 0.0f;
		   M(3,1) = 0.0f;
		   M(3,2) = 0.0f;
		   M(3,3) = 1.0f;

		   #undef M
		   return;
		}

		void MakeRotationX(float angle)
		{
			Identity();

			float s, c;
			_fpusincos(angle, s, c);

		    #define M(row,col)  m.v[(col<<2)+row]

		    M(1,1) = c;
		    M(1,2) =-s;
		    M(2,1) = s;
		    M(2,2) = c;

		    #undef M
		    return;
		}

		void MakeRotationY(float angle)
		{
			Identity();

			float s, c;
			_fpusincos(angle, s, c);

		    #define M(row,col)  m.v[(col<<2)+row]

		    M(0,0) = c;
		    M(0,2) = s;
		    M(2,0) =-s;
		    M(2,2) = c;

		    #undef M
		    return;
		}

		void MakeRotationZ(float angle)
		{
			Identity();

			float s, c;
			_fpusincos(angle, s, c);

		    #define M(row,col)  m.v[(col<<2)+row]

		    M(0,0) = c;
		    M(0,1) =-s;
		    M(1,0) = s;
		    M(1,1) = c;

		    #undef M
		    return;
		}
		
		friend int operator == (const c_MATRIX &a, const c_MATRIX &b)
		{
			for (int i = 0; i < 16; i++)
			{
				if (fabsf(a.m.v[i] - b.m.v[i]) > MATH_EPSILON)
					return FALSE;
			}

			return TRUE;
		}

		friend int operator != (const c_MATRIX &a, const c_MATRIX &b)
		{
			return (!(a == b));
		}

		void Lerp(const c_MATRIX &a, const c_MATRIX &b, float alpha);

		void LerpFromDecomposed(const c_MATRIX &a, const c_MATRIX &b, float alpha);

		BOOL Decompose(c_VECTOR *outScale, 
					   c_VECTOR *outRotate,
					   c_VECTOR *outTranslate) const;

		BOOL DecomposeInplace();

		void ComposeFromMatrix(const c_MATRIX &mat);

		void Compose(const c_VECTOR *inScale,
					 const c_VECTOR *inRotate,
					 const c_VECTOR *inTranslate);

		c_VECTOR ToQuat() const;

		BOOL Inverse();
};

extern const c_MATRIX gc_MATRIX_Identity;

///////////////////////////////////////////////////////////////////////////////
#endif // __MATRIX_H_INCLUDED__
