#include "misc.h"
#pragma  hdrstop

///////////////////////////////////////////////////////////////////////////////

static int         __mx__debug__line__;
static const char *__mx__debug__file__;

#define STRING_TO_CONSOLE

///////////////////////////////////////////////////////////////////////////////
#ifdef STRING_TO_CONSOLE
static HANDLE sHandleToConsole = INVALID_HANDLE_VALUE;
static void __FreeConsole()
{
 if (sHandleToConsole != INVALID_HANDLE_VALUE)
	 CloseHandle(sHandleToConsole);
 sHandleToConsole = INVALID_HANDLE_VALUE;
 FreeConsole();
 return;
}
#endif
///////////////////////////////////////////////////////////////////////////////
void __stdcall __OutputDebugString(const char *string)
{
#ifdef STRING_TO_CONSOLE
 static BOOL ConsoleAllocated = FALSE;
 if (ConsoleAllocated == FALSE)
 {
	 ConsoleAllocated = TRUE;

	 AllocConsole();
	 AttachConsole(GetCurrentProcessId());
	 atexit(__FreeConsole);
	 sHandleToConsole = CreateFile("CONOUT$", GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, 0, NULL);
	 if (sHandleToConsole != INVALID_HANDLE_VALUE)
		 SetStdHandle(STD_OUTPUT_HANDLE, sHandleToConsole);
 }

 if (sHandleToConsole != INVALID_HANDLE_VALUE)
 {
	 DWORD CharsOut = 0;
	 WriteConsole(sHandleToConsole, string, strlen(string), &CharsOut, NULL);
 }
#endif

 OutputDebugString(string);
 return;
}
///////////////////////////////////////////////////////////////////////////////
void __mx__debug__line_set__(int line)
{
 __mx__debug__line__ = line;
 return;
}
///////////////////////////////////////////////////////////////////////////////
void __mx__debug__file_set__(const char *file)
{
 __mx__debug__file__ = file;
 return;
}
///////////////////////////////////////////////////////////////////////////////
void __fucked_error (const char *format, ...)
{
 char tbffileline [MAX_FNAME * 2] = { 0 };
 char tbferror    [MAX_FNAME * 2] = { 0 };
 char tbffinal    [MAX_FNAME * 2] = { 0 };

 va_list arglist;
 sprintf_s(tbffileline, sizeof(tbffileline)/sizeof(tbffileline[0]), "[%s:%i] ERROR : ", __mx__debug__file__, __mx__debug__line__);

 if (format == NULL)
 {
	 sprintf_s(tbferror, sizeof(tbferror)/sizeof(tbferror[0]), "INTERNAL");
 }
 else
 {
	 va_start(arglist, format);
	 vsprintf_s(tbferror, sizeof(tbferror)/sizeof(tbferror[0]), format, arglist);
	 va_end(arglist);	 
 }

 sprintf_s(tbffinal, sizeof(tbffinal)/sizeof(tbffinal[0]), "%s %s \n", tbffileline, tbferror);

 __OutputDebugString(tbffinal);

 DebugBreak();

 return;
}
///////////////////////////////////////////////////////////////////////////////
void __mxLogger(const char *format, ...)
{
 char tmpbuffer[MAX_FNAME * 2];
 sprintf_s(tmpbuffer, sizeof(tmpbuffer) / sizeof(tmpbuffer[0]), "[%s:%i] : ", __mx__debug__file__, __mx__debug__line__);
 __OutputDebugString(tmpbuffer);

 if (format)
 {
	 va_list arglist;
	 va_start(arglist, format);
	 vsprintf_s(tmpbuffer, sizeof(tmpbuffer) / sizeof(tmpbuffer[0]), format, arglist);
	 va_end(arglist);	
	 __OutputDebugString(tmpbuffer);
 }

 __OutputDebugString("\n");

 return;
}
///////////////////////////////////////////////////////////////////////////////
void * __malloc_debug (size_t size, int line, const char *file)
{
 char *t = (char*)malloc(size);

 if (t == NULL) 
	 fucked_error("Cannot allocate %i memory [%s:%i] !", size, file, line);

 return t;
}
///////////////////////////////////////////////////////////////////////////////
void * __realloc_debug (void *ptr, size_t size, int line, const char *file)
{
 char *t = (char*)realloc(ptr, size);

 if (t == NULL) 
	 fucked_error("Cannot reassign %i memory [%s:%i] !", size, file, line);
 
 return t;
}
///////////////////////////////////////////////////////////////////////////////
void __free_debug (void *ptr, int line, const char *file)
{
 free(ptr);
 return;
}
///////////////////////////////////////////////////////////////////////////////
#pragma pack(1)
typedef struct tagTHREADNAME_INFO
{
  DWORD dwType;     // must be 0x1000
  LPCSTR szName;    // pointer to name (in user addr space)
  DWORD dwThreadID; // thread ID (-1=caller thread)
  DWORD dwFlags;    // reserved for future use, must be zero

} THREADNAME_INFO;
#pragma pack()

void __SetThreadName(DWORD dwThreadID, LPCSTR szThreadName)
{
  mxLogger("Thread %X named [%s]", (dwThreadID==0xFFFFFFFF) ? GetCurrentThreadId() : dwThreadID, szThreadName);

  THREADNAME_INFO info;
  {
    info.dwType		= 0x1000;
    info.szName		= szThreadName;
    info.dwThreadID = dwThreadID;
    info.dwFlags	= 0;
  }

  __try
  {
    RaiseException(0x406D1388, 0, sizeof(info)/sizeof(DWORD), (ULONG_PTR *)&info);
  }

  __except (EXCEPTION_CONTINUE_EXECUTION)
  {
  }

  return;
}
///////////////////////////////////////////////////////////////////////////////
void __SetThreadFPUCW()
{
#ifndef _X64
 const char *RoundName[] = 
 {
	 "Nearest",
	 "Down",
	 "Up",
	 "Chop",
 };

 const char *PrecisionName[] = 
 {
	 "64bit",
	 "53bit",
	 "24bit",
 };

 unsigned int OldWord = 0;

 _controlfp_s(&OldWord, 0, 0);
 _controlfp_s(&OldWord, _PC_24, _MCW_PC);
 _controlfp_s(&OldWord, 0, 0);

 mxLogger("FPUCW For thread %X is %X (Round: [%s], Precision: [%s])", GetCurrentThreadId(), 
																	  OldWord, 
																	  RoundName    [(OldWord & _MCW_RC) >> 8],
																	  PrecisionName[(OldWord & _MCW_PC) >> 16]);
#endif
 return;
}
///////////////////////////////////////////////////////////////////////////////
