#ifndef __SHAREDMISC_H_INCLUDED__
#define __SHAREDMISC_H_INCLUDED__

#define _WIN32_DCOM
#define _WIN32_WINNT 0x0500
#define _MFC_OVERRIDES_NEW
#define _NEW_

#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif

#ifdef __INTEL_COMPILER
#pragma warning( disable : 1684)
#endif
#pragma warning( disable : 4244)
#pragma warning( disable : 4267)
#pragma warning( disable : 4018)
#pragma warning( disable : 4200)
#pragma warning( disable : 4530)
#pragma warning( disable : 4700)

//
// STD includes
//
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <intrin.h>
#include <objbase.h>
#include <omp.h>
#include <process.h>

//
// STL includes
//
#include <algorithm>
#include <deque>
#include <functional>
#include <iterator>
#include <list>
#include <map>
#include <memory>
#include <numeric>
#include <queue>
#include <set>
#include <stack>
#include <utility>
#include <vector>

//
// DirectX
//
#include <d3d9.h>
#include <d3dx9.h>

//
// Cuda
//
#include <builtin_types.h>
#include <cuda_runtime_api.h>
#include <cuda_d3d9_interop.h>

#define TRUE						1
#define FALSE					 	0

#ifndef M_PI
#define M_PI                        3.14159265358979323846f
#endif

#ifndef SQR
#define SQR(x)                      ((x) * (x))
#endif

#ifndef DEG2RAD
#define DEG2RAD(a)					(((a) * M_PI) / 180.0f)
#endif

#ifndef RAD2DEG
#define RAD2DEG(a)					(((a)*360.0f) / (2.0f*M_PI)) 
#endif

#define malloc_debug(x)				__malloc_debug(x, __LINE__, __FILE__)
#define free_debug(x)				__free_debug(x, __LINE__, __FILE__)
#define realloc_debug(x, y)			__realloc_debug(x, y, __LINE__, __FILE__)
#define mxLogger					__mx__debug__line_set__(__LINE__), __mx__debug__file_set__(__FILE__), __mxLogger
#define fucked_error			    __mx__debug__line_set__(__LINE__), __mx__debug__file_set__(__FILE__), __fucked_error
#define MAX_FNAME                   512
#define MAX_FNAME_BONES				512
#define MAX_RESPECT                 100000000.0f
#define MIN_RESPECT                 (-MAX_RESPECT)
#define MATH_EPSILON                0.0001f
#define INTERPOLATORS_EPSI          0.005f

__forceinline float FLTlerp(float x, float y, float s)
{
 return x + (s * (y - x)); 
}

__forceinline float FLTsat(float v)
{
 if (v < 0.0) return 0.0;
 if (v > 1.0) return 1.0;
 return v;
}

__forceinline float FLTClamp(float x, float min, float max)
{
 if (x < min) x = min;
 if (x > max) x = max;
 return x;
}

__forceinline int FLTToInt(float val)
{
#ifdef __INTEL_COMPILER
 return _mm_cvtt_ss2si(_mm_load_ss(&val));
#else
#if _X64
 return _mm_cvtt_ss2si(_mm_load_ss(&val));
#else
 __asm cvttss2si eax, [val]
#endif
#endif
}

__forceinline __int64 FLTToInt64(float x)
{
#ifdef _X64
 return (__int64)x;
#else
 __int64 res;
 __asm
 {
	fld   x
	fistp qword ptr res
 }

 return res;
#endif
}

__forceinline float absf_(float r)
{
 *(long*)&r &= 0x7fffffff;
 return r;
}

__forceinline __m128 _mm_set_as_int(int a, int b, int c, int d)
{
 __m128 tmp;
 ((int*)&tmp)[0] = a;
 ((int*)&tmp)[1] = b;
 ((int*)&tmp)[2] = c;
 ((int*)&tmp)[3] = d;
 return tmp;
}

__forceinline __m128 _mm_set_as_flt(float a, float b, float c, float d)
{
 return _mm_setr_ps(a, b, c, d);
}

#ifndef __INTEL_COMPILER
#if defined (_MSC_VER) && (_MSC_VER < 1500)
__forceinline float _mm_cvtss_f32(__m128 reg)
{
 float v;
 _mm_store_ss(&v, reg);
 return v;
}
#endif
#endif

#define _mm_store_component(dest, reg, mask) { _mm_store_ss(&(dest), _mm_shuffle_ps((reg), (reg), _MM_SHUFFLE((mask), (mask), (mask), (mask)))); }

__forceinline float frand()
{
 return (float)(rand() - (RAND_MAX/2)) / ((float)RAND_MAX / 2.0f);
}

__forceinline int signum(float a)
{
 if (a < -MATH_EPSILON) return -1;
 if (a >  MATH_EPSILON) return  1;
 return 0;
}

__forceinline BYTE SignedFloatToBYTE(float a)
{
 a = max(-1.0f, min(1.0f, a));
 a = 0.5f + (a * 0.5f);
 return (BYTE)(a * 255.0f);
}

__forceinline float BYTEToSignedFloat(BYTE a)
{
 float v = (float)a;
 v /= 255.0f;
 v *= 2.0f;
 v -= 1.0f;
 return v;
}

__forceinline void _fpusincos(float f, float &s, float &c)
{
#ifdef _X64
 s = sinf(f);
 c = cosf(f);
#else
 float localCos, localSin;
 float local = f;

 _asm	fld		local
 _asm	fsincos
 _asm	fstp	localCos
 _asm	fstp	localSin

 c = localCos;
 s = localSin;
#endif
 return;
}

extern "C" void __mx__debug__line_set__(int line);
extern "C" void __mx__debug__file_set__(const char *file);
extern "C" void __fucked_error		   (const char *format, ...);
extern "C" void __mxLogger			   (const char *format, ...);
extern void    *__malloc_debug		   (size_t size, int line, const char *file);
extern void    *__realloc_debug		   (void *ptr, size_t size, int line, const char *file);
extern void     __free_debug		   (void *ptr, int line, const char *file);

#ifdef _X64
__forceinline int strEQ(const char *str1, const char *str2)
{
 return strcmp(str1, str2);
}
#else
inline int __declspec(naked) _fastcall strEQ(const char *str1, const char *str2)
{
 //
 // MMX string compare, about 50% faster than standard strcmp
 //
 __asm
 {
  push     edi
  push     esi
  push     ebx
  push     ecx
  mov      edi, ecx
  mov      esi, edx
  or	   edx, ecx
  test     edx, edx
  jnz      not_null_ptrs
  xor      eax, eax
  pop	   ecx
  pop      ebx
  pop      esi
  pop      edi
  ret
 
not_null_ptrs:
 
  test     edi, edi
  jz       not_eq
  test     esi, esi
  jz       not_eq
 
  pcmpeqb  mm7, mm7
 
compare:

  pxor	   mm2, mm2
  pxor	   mm3, mm3
  movq	   mm0, [esi]
  movq	   mm1, [edi]
  pcmpeqb  mm2, mm0
  pcmpeqb  mm3, mm1
  pcmpeqb  mm0, mm1
  pxor     mm0, mm7
  por      mm2, mm3
  por      mm2, mm0
  pmovmskb eax, mm2
  or       al, al  
  jnz      end     
  add	   esi, 8
  add	   edi, 8
  jmp      compare
 
end:
 
 emms
 bsf       edx, eax
 movzx     eax, byte ptr [edi+edx]
 movzx     ecx, byte ptr [esi+edx]
 sub       eax, ecx
 pop	   ecx
 pop	   ebx
 pop	   esi
 pop	   edi
 ret
 
not_eq:

  mov      eax, 1
  pop	   ecx
  pop      ebx
  pop      esi
  pop      edi
  ret
 }
}
#endif

inline void *operator new (size_t size)
{
 return malloc_debug(size);
}

inline void *operator new (size_t size, void *ptr)
{
 return ptr;
}

inline void operator delete (void *ptr)
{
 free_debug(ptr);
}

inline void operator delete (void *ptr, void *where)
{
 return;
}

inline void *operator new[] (size_t size)
{
 return malloc_debug(size);
}

inline void operator delete[] (void *ptr)
{
 free_debug(ptr);
}

#ifdef _X64
__forceinline float _absfmodf(const float a, const float b)
{
 return fabs(fmodf(a, b));
}
#else
inline float __declspec(naked) _fastcall _absfmodf(const float a, const float b)
{
 __asm
 {
	 fld dword ptr [esp+8] // b - st1
	 fld dword ptr [esp+4] // a - st0
	 fprem
	 fabs
	 ffree st(1)
	 ret 8
 }
}
#endif

#ifdef _X64
__forceinline float _absfmodf(const float a, const int b)
{
 return fabs(fmodf(a, (float)b));
}
#else
inline float __declspec(naked) _fastcall _absfmodf(const float a, const int b)
{
 __asm
 {
	 push ecx
	 fild dword ptr [esp]   // b - st1
	 fld  dword ptr [esp+8] // a - st0
	 fprem
	 pop  ecx
	 fabs
	 ffree st(1)
	 ret 4
 }
}
#endif

extern void __SetThreadName(DWORD dwThreadID, LPCSTR szThreadName);
extern void __SetThreadFPUCW();

///////////////////////////////////////////////////////////////////////////////
#endif // __SHAREDMISC_H_INCLUDED__
