#include "misc.h"
#include "vector.h"
#include "matrix.h"
#pragma  hdrstop

const c_VECTOR gc_VECTOR_Zero;

///////////////////////////////////////////////////////////////////////////////	
void c_VECTOR::Lerp(const c_VECTOR &a, const c_VECTOR &b, float t)
{
 float cosAngle = (a.v.x * b.v.x) +
				  (a.v.y * b.v.y) +
				  (a.v.z * b.v.z) +
				  (a.v.w * b.v.w); 

 float c1 = 1.0f - t;
 float c2 = t;

 //
 // Flips
 //
 if (cosAngle < 0.0)
     c1 = -c1;
  
 v.x = (c1 * a.v.x) + (c2 * b.v.x);
 v.y = (c1 * a.v.y) + (c2 * b.v.y);
 v.z = (c1 * a.v.z) + (c2 * b.v.z);
 v.w = (c1 * a.v.w) + (c2 * b.v.w);

 // 
 // Normalize quaternion
 //
 float s = 1.0f / sqrtf(SQR(v.x) + SQR(v.y) + SQR(v.z) + SQR(v.w));

 v.x *= s;
 v.y *= s;
 v.z *= s;
 v.w *= s;

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_VECTOR::QuatNormalize()
{
 float mag = SQR(v.x) + SQR(v.y) + SQR(v.z) + SQR(v.w);
 if (mag > MATH_EPSILON)
	 mag = 1.0f / sqrtf(mag);

 v.x *= mag;
 v.y *= mag;
 v.z *= mag;
 v.w *= mag;

 return;
}
///////////////////////////////////////////////////////////////////////////////	
void c_VECTOR::Slerp(const c_VECTOR &a, const c_VECTOR &b, float t)
{
 static const __m128 SlerpSgn = _mm_set_as_int(0x80000000, 0x80000000, 0x80000000, 0x80000000);
 static const __m128 SlerpAbs = _mm_set_as_int(0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF);
 static const __m128 SlerpEps = { 0.9999f,	        0.9999f,	    0.9999f,	    0.9999f         };
 static const __m128 SlerpOne = { 1.0000f,	        1.0000f,	    1.0000f,	    1.0000f         };
 static const __m128 SlerpHPi = { M_PI/2.0f,	    M_PI/2.0f,	    M_PI/2.0f,	    M_PI/2.0f       };
 static const __m128 SlerpAT1 = { 0.0028662257f,   -0.0161657367f,  0.0429096138f, -0.0752896400f   };
 static const __m128 SlerpAT2 = { 0.1065626393f,   -0.1420889944f,  0.1999355085f, -0.3333314528f   };
 static const __m128 SlerpAPS = {-4.0f/(M_PI*M_PI), 4.0f/M_PI,      0.0,			0.0				};
 static const __m128 SlerpAPP = { 0.2225f,			0.0,			0.0,			0.0				};

 __m128 _xmm0, _xmm1, _xmm2, _xmm3, _xmm4, _xmm5, _xmm6;
 
 __m128 veca	 = _mm_loadu_ps(a.v.v);
 __m128 vecb	 = _mm_loadu_ps(b.v.v);
 __m128 cosAngle = _mm_mul_ps(veca, vecb);
		cosAngle = _mm_add_ss(cosAngle, _mm_shuffle_ps(cosAngle, cosAngle, _MM_SHUFFLE(1, 1, 1, 1)));
		cosAngle = _mm_add_ss(cosAngle, _mm_shuffle_ps(cosAngle, cosAngle, _MM_SHUFFLE(3, 3, 3, 3)));
		cosAngle = _mm_add_ss(cosAngle, _mm_movehl_ps (cosAngle, cosAngle));

 _xmm0 = _mm_and_ps(cosAngle, SlerpAbs);

 if (_mm_comigt_ss(_xmm0, SlerpEps))
 {
	 _xmm0 = _mm_setzero_ps();
	 _xmm1 = _mm_load_ss   (&t);
	 _xmm2 = _mm_sub_ss    (SlerpOne, _xmm1);
	 _xmm3 = _mm_cmplt_ss  (cosAngle, _xmm0);
	 _xmm3 = _mm_and_ps    (_xmm3, SlerpSgn);
	 _xmm2 = _mm_xor_ps    (_xmm2, _xmm3);
	 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 0, 0));
	 _xmm2 = _mm_shuffle_ps(_xmm2, _xmm2, _MM_SHUFFLE(0, 0, 0, 0));

	 _mm_storeu_ps(v.v, _mm_add_ps(_mm_mul_ps(veca, _xmm2), _mm_mul_ps(vecb, _xmm1)));
	 return;
 }

 _xmm1 = _mm_mul_ss    (_xmm0, _xmm0);
 _xmm1 = _mm_sub_ss    (SlerpOne, _xmm1);
 _xmm2 = _mm_rsqrt_ss  (_xmm1);			 
 _xmm1 = _mm_mul_ss    (_xmm1, _xmm2);   
 
 //
 // _xmm0.x = atan2f(_xmm1.x, _xmm0.x);
 //
 _xmm4 = _mm_cmpgt_ss  (_xmm1, _xmm0);
 _xmm5 = _mm_max_ss    (_xmm1, _xmm0);
 _xmm6 = _mm_min_ss    (_xmm1, _xmm0);
 _xmm5 = _mm_rcp_ss    (_xmm5);
 _xmm6 = _mm_mul_ss    (_xmm6, _xmm5);
 _xmm6 = _mm_xor_ps    (_xmm6, _mm_and_ps(_xmm4, SlerpSgn));
 _xmm4 = _mm_and_ps    (_xmm4, SlerpHPi);
 _xmm0 = _mm_mul_ss    (_xmm6, _xmm6);
 _xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm1 = _mm_mul_ps    (_xmm0, SlerpAT1);
 _xmm0 = _mm_mul_ps    (_xmm0, SlerpAT2);
 _xmm0 = _mm_add_ps    (_xmm0, _xmm1);
 _xmm0 = _mm_add_ss    (_xmm0, _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1)));
 _xmm0 = _mm_add_ss    (_xmm0, _mm_movehl_ps (_xmm0, _xmm0));
 _xmm0 = _mm_add_ss    (_xmm0, _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 3)));
 _xmm0 = _mm_add_ss    (_xmm0, SlerpOne);
 _xmm0 = _mm_mul_ss	   (_xmm0, _xmm6);
 _xmm0 = _mm_add_ss	   (_xmm0, _xmm4);

 _xmm5 = _mm_load_ss   (&t);
 _xmm1 = _mm_mul_ss    (_mm_sub_ss(SlerpOne, _xmm5), _xmm0);

 //
 // _xmm3.x = sinf(_xmm1.x);
 //
 _xmm3 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm4 = _mm_mul_ps    (_xmm3, SlerpAPS);
 _xmm3 = _mm_and_ps    (_xmm3, SlerpAbs);
 _xmm6 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm4 = _mm_mul_ss	   (_xmm4, _xmm3);
 _xmm4 = _mm_add_ss	   (_xmm4, _xmm6);
 _xmm3 = _mm_and_ps	   (_xmm4, SlerpAbs);
 _xmm3 = _mm_mul_ss	   (_xmm3, _xmm4);
 _xmm3 = _mm_sub_ss    (_xmm3, _xmm4);
 _xmm3 = _mm_mul_ss    (_xmm3, SlerpAPP);
 _xmm3 = _mm_add_ss    (_xmm3, _xmm4);

 _xmm3 = _mm_mul_ss	   (_xmm3, _xmm2);
 _xmm0 = _mm_mul_ss	   (_xmm0, _xmm5);

 //
 // _xmm6.x = sinf(_xmm0.x);
 //
 _xmm6 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm4 = _mm_mul_ps    (_xmm6, SlerpAPS);
 _xmm6 = _mm_and_ps    (_xmm6, SlerpAbs);
 _xmm1 = _mm_shuffle_ps(_xmm4, _xmm4, _MM_SHUFFLE(1, 1, 1, 1));
 _xmm4 = _mm_mul_ss    (_xmm4, _xmm6);
 _xmm4 = _mm_add_ss    (_xmm4, _xmm1);
 _xmm6 = _mm_and_ps    (_xmm4, SlerpAbs);
 _xmm6 = _mm_mul_ss    (_xmm6, _xmm4);
 _xmm6 = _mm_sub_ss    (_xmm6, _xmm4);
 _xmm6 = _mm_mul_ss	   (_xmm6, SlerpAPP);
 _xmm6 = _mm_add_ss    (_xmm6, _xmm4);

 _xmm1 = _mm_setzero_ps();
 _xmm6 = _mm_mul_ss	   (_xmm6, _xmm2);
 _xmm0 = _mm_cmplt_ss  (cosAngle, _xmm1);
 _xmm0 = _mm_and_ps	   (_xmm0, SlerpSgn);
 _xmm6 = _mm_shuffle_ps(_xmm6, _xmm6, _MM_SHUFFLE(0, 0, 0, 0));
 _xmm3 = _mm_xor_ps    (_xmm3, _xmm0);
 _xmm3 = _mm_shuffle_ps(_xmm3, _xmm3, _MM_SHUFFLE(0, 0, 0, 0));

 _xmm0 = _mm_add_ps(_mm_mul_ps(veca, _xmm3), _mm_mul_ps(vecb, _xmm6));

 //
 // Need to normalize result
 //
 _xmm1 = _mm_mul_ps    (_xmm0, _xmm0);
 _xmm1 = _mm_add_ss    (_xmm1, _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1)));
 _xmm1 = _mm_add_ss    (_xmm1, _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(2, 2, 2, 2)));
 _xmm1 = _mm_add_ss    (_xmm1, _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(3, 3, 3, 3)));
 _xmm1 = _mm_rsqrt_ss  (_xmm1);
 _xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 0, 0));

 _mm_storeu_ps(v.v, _mm_mul_ps(_xmm0, _xmm1));

 return;
}
///////////////////////////////////////////////////////////////////////////////	
