#ifndef __VECTOR_H_INCLUDED__
#define __VECTOR_H_INCLUDED__

///////////////////////////////////////////////////////////////////////////////
#pragma pack(1)
typedef struct _c_VECTOR_STR
{
 union {
        struct { float x, y, z, w;						};
        struct { float v[4];							};
		struct { int   ix, iy, iz, iw;					};
       };
}c_VECTOR_STR;
#pragma pack()

class c_VECTOR
{
   public:

        c_VECTOR_STR  v;

		c_VECTOR(__m128 reg)
		{
			_mm_storeu_ps(v.v, reg);
		}

		c_VECTOR(float x, float y, float z, float w)
		{
			v.x = x; v.y = y; v.z = z; v.w = w;
		}

		c_VECTOR(float x, float y, float z, DWORD w)
		{
			v.x = x; v.y = y; v.z = z; v.iw = w;
		}

		c_VECTOR()
		{
			_mm_storeu_ps(v.v, _mm_setzero_ps());
		}

		void Zero()
		{
			_mm_storeu_ps(v.v, _mm_setzero_ps());
		}

		void One()
		{
			_mm_storeu_ps(v.v, _mm_set_ps1(1.0));
		}

		void SetMinimum()
		{
			_mm_storeu_ps(v.v, _mm_set_ps1(MIN_RESPECT));
		}

		void SetMaximum()
		{
			_mm_storeu_ps(v.v, _mm_set_ps1(MAX_RESPECT));
		}

		int MaxComponentXZY() const
		{
			if (v.x > v.y)
			{
				if (v.x > v.z)
					return 0;
			}
			else
			{
				if (v.y > v.z)
					return 1;
			}

			return 2;
		}

		c_VECTOR operator * (const float &oth) const
		{
			return _mm_mul_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
		}

		c_VECTOR operator / (const float &oth) const
		{
			return _mm_div_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
		}

		c_VECTOR operator + (const float &oth) const
		{
			return _mm_add_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
		}

		c_VECTOR operator - (const float &oth) const
		{
			return _mm_sub_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
		}

		c_VECTOR operator + (const c_VECTOR &oth) const
		{
			return _mm_add_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
		}

		c_VECTOR operator - (const c_VECTOR &oth) const
		{
			return _mm_sub_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
		}

		c_VECTOR operator * (const c_VECTOR &oth) const
		{
			return _mm_mul_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
		}

		c_VECTOR operator / (const c_VECTOR &oth) const
		{
			return _mm_div_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
		}

		c_VECTOR operator += (const c_VECTOR &oth)
		{
			*this = _mm_add_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
			return *this;
		}

		c_VECTOR operator -= (const c_VECTOR &oth)
		{
			*this = _mm_sub_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
			return *this;
		}

		c_VECTOR operator *= (const c_VECTOR &oth)
		{
			*this = _mm_mul_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
			return *this;
		}

		c_VECTOR operator /= (const c_VECTOR &oth)
		{
			*this = _mm_div_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(oth.v.v));
			return *this;
		}

		c_VECTOR operator += (float oth)
		{
			*this = _mm_add_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
			return *this;
		}

		c_VECTOR operator -= (float oth)
		{
			*this = _mm_sub_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
			return *this;
		}

		c_VECTOR operator *= (float oth)
		{
			*this = _mm_mul_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
			return *this;
		}

		c_VECTOR operator /= (float oth)
		{
			*this = _mm_div_ps(_mm_loadu_ps(v.v), _mm_set_ps1(oth));
			return *this;
		}

		friend int operator == (const c_VECTOR &a, const c_VECTOR &b)
		{
			__m128 _xmm0 = _mm_sub_ps(_mm_loadu_ps(a.v.v), _mm_loadu_ps(b.v.v));
				   _xmm0 = _mm_max_ps(_xmm0, _mm_sub_ps(_mm_setzero_ps(), _xmm0));
				   _xmm0 = _mm_cmpgt_ps(_xmm0, _mm_set1_ps(MATH_EPSILON));

		    if (_mm_movemask_ps(_xmm0)) 
				return FALSE;

			return TRUE;
		}

		friend int operator != (const c_VECTOR &a, const c_VECTOR &b)
		{
			return (!(a == b));
		}

		c_VECTOR Cross(const c_VECTOR &b) const 
		{
			__m128 _xmm0, _xmm1, _xmm2, _xmm3;

			_xmm0 = _mm_loadu_ps(v.v);
			_xmm1 = _mm_loadu_ps(b.v.v);

			_xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 0, 2, 1));
			_xmm3 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(3, 1, 0, 2));

			_xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 1, 0, 2));
			_xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(3, 0, 2, 1));

			_xmm2 = _mm_mul_ps(_xmm2, _xmm3);
			_xmm0 = _mm_mul_ps(_xmm0, _xmm1);

			return _mm_sub_ps(_xmm2, _xmm0);
		}

		c_VECTOR UnitCross(const c_VECTOR &b) const 
		{
			__m128 _xmm0, _xmm1, _xmm2, _xmm3;

			_xmm0 = _mm_loadu_ps(v.v);
			_xmm1 = _mm_loadu_ps(b.v.v);

			_xmm2 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 0, 2, 1));
			_xmm3 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(3, 1, 0, 2));

			_xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(2, 1, 0, 2));
			_xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(3, 0, 2, 1));

			_xmm2 = _mm_mul_ps(_xmm2, _xmm3);
			_xmm0 = _mm_mul_ps(_xmm0, _xmm1);

			_xmm2 = _mm_sub_ps(_xmm2, _xmm0);

			_xmm0 = _mm_mul_ps(_xmm2, _xmm2);
			_xmm1 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm3 = _mm_movehl_ps(_xmm3, _xmm0);
			_xmm0 = _mm_add_ss(_xmm0, _xmm3);
			_xmm3 = _mm_setzero_ps();
			_xmm0 = _mm_add_ss(_xmm0, _xmm1);

			if (_mm_comieq_ss(_xmm1, _xmm3))
				return _xmm3;

			_xmm0 = _mm_rsqrt_ss(_xmm0);
			_xmm0 = _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(0, 0, 0, 0));

			return _mm_mul_ps(_xmm0, _xmm2);
		}

		float Dot3(const c_VECTOR &b) const
		{
			__m128 _xmm0 = _mm_mul_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(b.v.v));
			       _xmm0 = _mm_add_ss(_xmm0, _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1)));
				   _xmm0 = _mm_add_ss(_xmm0, _mm_movehl_ps(_xmm0, _xmm0));
		    return _mm_cvtss_f32(_xmm0);
		}

		float Dot4(const c_VECTOR &b) const
		{
			__m128 _xmm0 = _mm_mul_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(b.v.v));
			       _xmm0 = _mm_add_ss(_xmm0, _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(1, 1, 1, 1)));
				   _xmm0 = _mm_add_ss(_xmm0, _mm_movehl_ps (_xmm0, _xmm0));
				   _xmm0 = _mm_add_ss(_xmm0, _mm_shuffle_ps(_xmm0, _xmm0, _MM_SHUFFLE(3, 3, 3, 3)));
		    return _mm_cvtss_f32(_xmm0);
		}

		c_VECTOR Normalize3D()
		{
			__m128 _xmm1, _xmm2, _xmm3, _xmm4;

			_xmm4 = _mm_loadu_ps(v.v);
			_xmm1 = _mm_mul_ps(_xmm4, _xmm4);
			_xmm3 = _mm_movehl_ps(_xmm1, _xmm1);
			_xmm2 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm1 = _mm_add_ss(_xmm1, _xmm3);
			_xmm3 = _mm_setzero_ps();
			_xmm1 = _mm_add_ss(_xmm1, _xmm2);

			if (_mm_comieq_ss(_xmm1, _xmm3))
			{
				_mm_storeu_ps(v.v, _xmm3);
				return *this;
			}

			_xmm1 = _mm_rsqrt_ss(_xmm1);
			_xmm1 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(0, 0, 0, 0));
			_xmm4 = _mm_mul_ps(_xmm4, _xmm1);
			_mm_storeu_ps(v.v, _xmm4);

			return *this;
		}

		float Len3() const 
		{
			__m128 _xmm1, _xmm2, _xmm3, _xmm4;

			_xmm4 = _mm_loadu_ps(v.v);
			_xmm1 = _mm_mul_ps(_xmm4, _xmm4);
			_xmm3 = _mm_movehl_ps(_xmm1, _xmm1);
			_xmm2 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm1 = _mm_add_ss(_xmm1, _xmm3);
			_xmm3 = _mm_setzero_ps();
			_xmm1 = _mm_add_ss(_xmm1, _xmm2);

			if (_mm_comieq_ss(_xmm1, _xmm3))
				return 0.0f;

			return _mm_cvtss_f32(_mm_sqrt_ss(_xmm1));
		}

		float Len3SQR() const 
		{
			__m128 _xmm1, _xmm2, _xmm3, _xmm4;

			_xmm4 = _mm_loadu_ps(v.v);
			_xmm1 = _mm_mul_ps(_xmm4, _xmm4);
			_xmm3 = _mm_movehl_ps(_xmm1, _xmm1);
			_xmm2 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm1 = _mm_add_ss(_xmm1, _xmm3);
			_xmm3 = _mm_setzero_ps();
			_xmm1 = _mm_add_ss(_xmm1, _xmm2);

			return _mm_cvtss_f32(_xmm1);
		}

		float Distance(const c_VECTOR &other) const 
		{
			__m128 _xmm1, _xmm2, _xmm3, _xmm4;

			_xmm4 = _mm_sub_ps(_mm_loadu_ps(v.v), _mm_loadu_ps(other.v.v));
			_xmm1 = _mm_mul_ps(_xmm4, _xmm4);
			_xmm3 = _mm_movehl_ps(_xmm1, _xmm1);
			_xmm2 = _mm_shuffle_ps(_xmm1, _xmm1, _MM_SHUFFLE(1, 1, 1, 1));
			_xmm1 = _mm_add_ss(_xmm1, _xmm3);
			_xmm3 = _mm_setzero_ps();
			_xmm1 = _mm_add_ss(_xmm1, _xmm2);

			if (_mm_comieq_ss(_xmm1, _xmm3))
				return 0.0;

			return _mm_cvtss_f32(_mm_sqrt_ss(_xmm1));
		}

		void Interpolate(const c_VECTOR &v1, const c_VECTOR &v2, float factor)
		{
			__m128 _xmm0 = _mm_loadu_ps(v1.v.v);
			__m128 _xmm1 = _mm_loadu_ps(v2.v.v);
			*this = _mm_add_ps(_xmm0, _mm_mul_ps(_mm_sub_ps(_xmm1, _xmm0), _mm_set_ps1(factor)));
		}

		void RotateX(float angle)
		{
			float siny, cosy;
			 _fpusincos(angle, siny, cosy);

			 float poz_y = v.y;
			 float poz_z = v.z;

			 v.y = (poz_y * cosy) - (poz_z * siny);
			 v.z = (poz_y * siny) + (poz_z * cosy);

			 return;
		}
		
		void RotateY(float angle)
		{
			float siny, cosy;
			 _fpusincos(angle, siny, cosy);

			 float poz_x = v.x;
			 float poz_z = v.z;

			 v.x = (poz_x * cosy) - (poz_z * siny);
			 v.z = (poz_x * siny) + (poz_z * cosy);

			 return; 
		}

		void RotateZ(float angle)
		{
			float siny, cosy;
			 _fpusincos(angle, siny, cosy);

			 float poz_x = v.x;
			 float poz_y = v.y;

			 v.x = (poz_x * cosy) - (poz_y * siny);
			 v.y = (poz_x * siny) + (poz_y * cosy);

			 return;
		}

		//
		// Spherical interpolation of quaternions
		//
		void Slerp(const c_VECTOR &a, const c_VECTOR &b, float alpha);

		//
		// Linear (a bit faster, but no const velocity in rotation over alpha time) interpolation of quaternion
		//
		void Lerp(const c_VECTOR &a, const c_VECTOR &b, float alpha);

		//
		// Quat normalization
		//
		void QuatNormalize();

		c_VECTOR operator-() const 
		{
			return c_VECTOR(-v.x, -v.y, -v.z, -v.w); 
		}

		void FromPackedDWORD_XYZ(DWORD pck)
		{
			v.x = BYTEToSignedFloat((BYTE)(pck >> 16));
			v.y = BYTEToSignedFloat((BYTE)(pck >>  8));
			v.z = BYTEToSignedFloat((BYTE)(pck      ));
			v.w = 0;
			return;
		}

		DWORD ToPackedDWORD_XYZ() const
		{
			DWORD res = 0;

				  res |= ((DWORD)SignedFloatToBYTE(v.x)) << 16;
				  res |= ((DWORD)SignedFloatToBYTE(v.y)) <<  8;
				  res |= ((DWORD)SignedFloatToBYTE(v.z))      ;

		    return res;
		}

		void Clamp()
		{
			__m128 t = _mm_loadu_ps(v.v);
			t = _mm_max_ps(t, _mm_setzero_ps());
			t = _mm_min_ps(t, _mm_set_ps1(1.0f));
			_mm_storeu_ps(v.v, t);
			return;
		}

		DWORD ToRGB() const
		{
			DWORD res = 0xFF000000;

			res |= (DWORD)(v.x * 255.0f) << 16;
			res |= (DWORD)(v.y * 255.0f) <<  8;
			res |= (DWORD)(v.z * 255.0f)      ;

			return res;
		}

		static c_VECTOR MinAll(const c_VECTOR &a, const c_VECTOR &b)
		{
			return _mm_min_ps(_mm_loadu_ps(a.v.v), _mm_loadu_ps(b.v.v));
		}

		static c_VECTOR MaxAll(const c_VECTOR &a, const c_VECTOR &b)
		{
			return _mm_max_ps(_mm_loadu_ps(a.v.v), _mm_loadu_ps(b.v.v));
		}
};

extern const c_VECTOR gc_VECTOR_Zero;

///////////////////////////////////////////////////////////////////////////////
#endif // __VECTOR_H_INCLUDED__
